#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

from abc import abstractmethod
import math
import pprint
import random

import mlflow
import numpy as np
import torch

from openqa_classifier import make_surrounded


def print_metrics(metrics, metric_label):
    primary_metrics = ["accuracy", "precision", "recall", "f1"]
    print("")
    print(
        "********************************************************************************"
    )
    print(make_surrounded("{} Metrics".format(metric_label)))
    print(
        "********************************************************************************"
    )
    print("")
    for metric in primary_metrics:
        label_string = "{}:".format(metric.title())
        print("{:<20} {:.4f}".format(label_string, metrics[metric]))
    print("")
    for k in metrics.keys():
        if k in primary_metrics:
            continue
        label_string = "{}:".format(k)
        metric = metrics[k]
        if isinstance(metric, torch.Tensor):
            # tensor_string = "{} ({}, {})".format(metric, metric.shape, metric.dtype)
            tensor_string = "{}".format(metric)
            print("{:<20} {:}".format(label_string, tensor_string))
        else:
            print("{:<20} {}".format(label_string, metric))
    print("")


def calculate_mean_metrics(metrics):
    mean_metrics = {}
    for k in metrics[0]:
        metric_values = []
        for fold_metric in metrics:
            metric_values.append(fold_metric[k])
        mean_metrics[k] = np.mean(metric_values)
    return mean_metrics


class OpenQAClassifier(object):
    def __init__(self, base_dir, config, output_dir):
        self.base_dir = base_dir
        self.config = config
        self.output_dir = output_dir
        self.device = "cpu"
        self.model = None
        self.validation_outputs = []
        self.start_time = None
        self.reporting_interval = 25
        self.k_folds = 0
        self.num_data_workers = 6
        self.mlflow_run_name = "defaultname-thisisabug"

        self.dataset = None
        self.train_dataloader = None
        self.val_dataloader = None
        self.test_dataloader = None

        self.test_metrics = []
        self.validation_metrics = []

    def log_mlflow_metrics(self, metrics_dict, prefix=""):
        mlflow_metric_names = [
            "accuracy",
            "precision",
            "recall",
            "f1",
            "items",
            "label_sum",
            "output_sum",
            "prediction_sum",
            "runtime",
        ]
        mlflow_metrics = dict(
            ("{}{}".format(prefix, k), metrics_dict[k])
            for k in mlflow_metric_names
            if k in metrics_dict
        )

        # we need to convert some of the metrics into scalars
        for name in ["label_sum", "output_sum", "prediction_sum"]:
            prefixed_name = "{}{}".format(prefix, name)
            datum = metrics_dict[name]
            # we're assuming that the tensor is effectively a scalar, set the value as -1 if our assumption is wrong
            if datum.dim() != 0:
                mlflow_metrics[prefixed_name] = -1.0
            else:
                mlflow_metrics[prefixed_name] = datum.item()

        mlflow.log_metrics(mlflow_metrics)

    def _setup_mlflow_experiment(self):
        mlflow_experiment_name = "OpenQA Reschedule Classifier"
        if "mlflow_experiment_name" in self.config.keys():
            mlflow_experiment_name = self.config["mlflow_experiment_name"]
        self.ml_experiment = mlflow.set_experiment(mlflow_experiment_name)

    def _setup_mlflow_run(self, kfold_num):
        self.mlflow_run_name = "openqa-{}-cross-{}-{}".format(
            self.config["iteration_id"], kfold_num, self.k_folds
        )
        self.ml_run = mlflow.start_run(
            run_name=self.mlflow_run_name,
            experiment_id=self.ml_experiment.experiment_id,
        )
        mlflow.log_params(self.config)
        mlflow.log_param("kfold_num", kfold_num)

    @abstractmethod
    def prepare_data(self, use_portion):
        # set self.dataset to the loaded torch dataset
        pass

    def _create_dataloaders(self, test_idx, train_idx):
        # need to split training set into train and validation dataloaders
        real_train_idx, val_idx = self._split_array(train_idx, self.train_val_split)
        self.train_dataloader = torch.utils.data.DataLoader(
            dataset=self.dataset,
            batch_size=self.config["batch_size"],
            num_workers=self.num_data_workers,
            sampler=torch.utils.data.SubsetRandomSampler(real_train_idx),
        )
        self.val_dataloader = torch.utils.data.DataLoader(
            dataset=self.dataset,
            batch_size=self.config["batch_size"],
            num_workers=self.num_data_workers,
            sampler=torch.utils.data.SubsetRandomSampler(val_idx),
        )
        self.test_dataloader = torch.utils.data.DataLoader(
            dataset=self.dataset,
            batch_size=self.config["batch_size"],
            num_workers=self.num_data_workers,
            sampler=torch.utils.data.SubsetRandomSampler(test_idx),
        )

    def _do_run_setup(self):
        print("Using Config:")

        pprint.pprint(self.config)
        print("")

        # set random seeds
        random.seed(self.config["random_seed"])
        torch.manual_seed(self.config["random_seed"])
        np.random.seed(self.config["random_seed"])

        if not self.config["force_cpu"]:
            self.device = torch.device("cuda:0") if torch.cuda.is_available() else "cpu"

        # this is used to split train set into train/val sets
        self.split_rng = np.random.default_rng(self.config["random_seed"])

        self._prepare_model()
        self.prepare_data(self.data_use_portion)
        self._setup_mlflow_experiment()

    def _split_array(self, array, fraction):
        array_len = len(array)
        first_size = math.ceil(array_len * fraction)

        shuffled_array = self.split_rng.permutation(array)

        return shuffled_array[:first_size], shuffled_array[first_size:]
