#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import copy
import os

import datasets as hf_datasets
import numpy as np
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
import torch
from torchvision import transforms
from transformers import TrainingArguments, Trainer
from transformers import AutoImageProcessor, AutoModelForImageClassification

from openqa_classifier.classifiers import OpenQAClassifier, calculate_mean_metrics

"""
Use pre-trained vision transformer model for openqa result classification.

Based on: https://github.com/NielsRogge/Transformers-Tutorials/blob/master/VisionTransformer/Fine_tuning_the_Vision_Transformer_on_CIFAR_10_with_the_%F0%9F%A4%97_Trainer.ipynb
"""


def compute_metrics(eval_pred):
    predictions, labels = eval_pred
    predictions = np.argmax(predictions, axis=1)
    return dict(
        accuracy=accuracy_score(predictions, labels),
        precision=precision_score(predictions, labels),
        recall=recall_score(predictions, labels),
        f1=f1_score(predictions, labels),
    )


def collate_fn(examples):
    pixel_values = torch.stack([example["pixel_values"] for example in examples])
    labels = torch.tensor([example["label"] for example in examples])
    return {"pixel_values": pixel_values, "labels": labels}


class OpenQAClassifierVIT(OpenQAClassifier):
    def __init__(self, base_dir, config, output_dir):
        super().__init__(base_dir, config, output_dir)
        self.pretrained_name = config["model_name"]

        self.id2label = None
        self.label2id = None

        self.train_dataset = None
        self.val_dataset = None
        self.test_dataset = None

        self.training_metric = "accuracy"

        self.processor = None
        self.training_args = None
        self.trainer = None

    def run_single_iteration(self, use_portion=1.0):
        self._prepare_model()
        self.prep_training()
        val_metrics = self.run_training()
        test_metrics = self.run_eval()
        return {"validation": val_metrics, "test": test_metrics}

    def prepare_data(self, use_portion):
        if "dataset_path" in self.config.keys():
            screenshot_dataset_path = self.config["dataset_path"]
        else:
            raise Exception("dataset_path not in run config")

        self.processor = AutoImageProcessor.from_pretrained(self.pretrained_name)

        image_mean, image_std = self.processor.image_mean, self.processor.image_std

        size = (
            self.processor.size["shortest_edge"]
            if "shortest_edge" in self.processor.size
            else (self.processor.size["height"], self.processor.size["width"])
        )

        dataset_transforms = transforms.Compose(
            [
                transforms.Resize(size),
                transforms.CenterCrop(size),
                transforms.ToTensor(),
                transforms.Normalize(mean=image_mean, std=image_std),
            ]
        )

        def apply_dataset_transforms(data):
            data["pixel_values"] = [
                dataset_transforms(image.convert("RGB")) for image in data["image"]
            ]
            return data

        dataset = hf_datasets.load_dataset(
            "imagefolder", data_dir=screenshot_dataset_path
        )

        split_ds = dataset["train"].train_test_split(test_size=0.2)
        fulltrain_ds = split_ds["train"]
        fulltrain_split = fulltrain_ds.train_test_split(test_size=0.1)

        self.test_dataset = split_ds["test"]
        self.train_dataset = fulltrain_split["train"]
        self.val_dataset = fulltrain_split["test"]

        self.train_dataset.set_transform(apply_dataset_transforms)
        self.test_dataset.set_transform(apply_dataset_transforms)
        self.val_dataset.set_transform(apply_dataset_transforms)

        self.id2label = {
            key: label
            for key, label in enumerate(self.train_dataset.features["label"].names)
        }
        self.label2id = {label: key for key, label in self.id2label.items()}

    def _prepare_model(self):
        self.model = AutoModelForImageClassification.from_pretrained(
            self.pretrained_name,
            num_labels=2,
            id2label=self.id2label,
            label2id=self.label2id,
            ignore_mismatched_sizes=True,
        )

    def prep_training(self, iteration=1):
        # the hf trainer looks for an environment variable to determine the mlflow experiment name
        os.environ["MLFLOW_EXPERIMENT_NAME"] = self.config["mlflow_experiment_name"]

        run_name = "{}-{}-{}".format(
            self.config["model_class"], self.config["iteration_id"], iteration
        )

        hf_log_dir = os.path.join(self.output_dir, "hf_training")

        if self.model is None:
            self._prepare_model()

        self.training_args = TrainingArguments(
            hf_log_dir,
            do_train=True,
            save_strategy="no",
            evaluation_strategy="epoch",
            learning_rate=2e-5,
            per_device_train_batch_size=self.config["batch_size"],
            per_device_eval_batch_size=self.config["batch_size"],
            num_train_epochs=self.config["num_epochs"],
            weight_decay=0.01,
            load_best_model_at_end=False,
            metric_for_best_model=self.training_metric,
            logging_dir=hf_log_dir,
            remove_unused_columns=False,
            seed=self.config["random_seed"],
            run_name=run_name,
            use_cpu=self.config["force_cpu"],
        )

        self.trainer = Trainer(
            self.model,
            self.training_args,
            train_dataset=self.train_dataset,
            eval_dataset=self.val_dataset,
            data_collator=collate_fn,
            compute_metrics=compute_metrics,
            tokenizer=self.processor,
        )

    def run_training(self):
        training_output = self.trainer.train()
        training_history = copy.deepcopy(self.trainer.state.log_history)
        # eval_loss, eval_accuracy, eval_precision, eval_recall, eval_f1, eval_runtime, epoch (float), step, eval_samples_per_second, eval_steps_per_second
        # epoch: 5.0 - train_runtime, train_samples_per_second, train_steps_per_second, total_flos, train_loss, epoch, step
        # epoch: 3.6 - loss, learning_rate, epoch, step
        # training history dicts will have integer epoch and eval_* for metrics
        run_metrics = {}
        just_metrics = []
        for entry in training_history:
            if not entry["epoch"].is_integer():
                continue

            if "eval_accuracy" in entry.keys():
                metrics = {}
                for key in entry.keys():
                    if not key.startswith("eval"):
                        continue
                    metrics[key[5:]] = entry[key]
                epoch_key = "epoch_{}".format(int(entry["epoch"]))
                run_metrics[epoch_key] = metrics
                just_metrics.append(metrics)

            elif "train_runtime" in entry.keys():
                for key in entry.keys():
                    if key.startswith("train"):
                        run_metrics[key] = entry[key]
        run_metrics["mean"] = calculate_mean_metrics(just_metrics)
        self.validation_metrics.append(run_metrics)

        return run_metrics

    def run_eval(self):
        eval_output = self.trainer.predict(self.test_dataset)
        metrics = eval_output.metrics
        self.test_metrics.append(metrics)
        return metrics
