#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

from datetime import datetime
import math
import os

from sklearn.metrics import precision_score, recall_score, f1_score, accuracy_score
from sklearn.model_selection import KFold

import mlflow

import torch
import torch.nn as nn
from torch.utils.data import Dataset
from torchvision import datasets, transforms

from openqa_classifier import DATE_FORMAT, RANDOM_SEED, make_surrounded
from openqa_classifier.classifiers import (
    print_metrics,
    OpenQAClassifier,
    calculate_mean_metrics,
)
from openqa_classifier.models import resnet, vgg, vision_transformer, cnn_3_layer

BATCH_SIZE = 16
NUM_EPOCHS = 2
INIT_LR = 1e-3
TEST_SPLIT = 0.2
TRAIN_SPLIT = 1 - TEST_SPLIT
VAL_SPLIT = 0.2
TRAIN_VAL_SPLIT = 1 - VAL_SPLIT


def classify_images(
    base_dir, num_epochs=NUM_EPOCHS, use_portion=1.0, force_cpu=False, dump_values=False
):
    # this was the best performing config after doing a grid search
    model_config = {
        "cnn1_input_channels": 3,
        "cnn1_output_channels": 10,
        "cnn2_output_channels": 15,
        "cnn3_output_channels": 12,
        "cnn_stride": 1,
        "kernel_size": 3,
        "layer_1_size": 100,
        "layer_end_size": 1,
        "max_pool_kernel": 1,
        "max_pool_stride": 2,
        "image_height": 1536,
        "image_width": 2048,
        "force_cpu": force_cpu,
        "random_seed": RANDOM_SEED,
        "batch_size": BATCH_SIZE,
        "num_epochs": num_epochs,
        #'cnn_padding': 0,
        #'cnn_dilation': 1,
    }


class OpenQATorchClassifier(OpenQAClassifier):
    def __init__(self, base_dir, config, output_dir):
        super().__init__(base_dir, config, output_dir)

        self.required_config_keys = None
        self.data_transforms = None

        self.data_use_portion = 1.0

        # the fraction of data to use for train (remainder for test)
        self.train_test_split = 0.8

        # the fraction of training set to use for training (remainder for validation)
        self.train_val_split = 0.8

    def _validate_config(self):
        """This runs some very rudimentary checks on config, mostly to see if extra stuff we require is in there. If
        model params are missing ... you're going to be in trouble anyways"""
        if "iteration_id" not in self.config.keys():
            print("WARNING: config does not have iteration ID. Setting to default")
            self.config["iteration_id"] = 0

        missing_config_keys = []
        for key in self.required_config_keys:
            if key not in self.config.keys():
                missing_config_keys.append(key)

        if len(missing_config_keys) > 0:
            missing_keys = ",".join(missing_config_keys)
            raise Exception(
                "Required key(s) not found in config: {}".format(missing_keys)
            )

    def _prepare_model(self):
        model_name = "cnn_3_layer"
        if "model_name" in self.config.keys():
            model_name = self.config["model_name"]

        if model_name == "cnn_3_layer":
            self.model = cnn_3_layer.OpenqaRescueduleClassifier(self.config)
            self.criterion = cnn_3_layer.criterion
            self.required_config_keys = cnn_3_layer.required_config_keys
            self.data_transforms = cnn_3_layer.data_transforms
        elif model_name in resnet.model_names:
            self.model = resnet.get_model(model_name)
            self.criterion = resnet.criterion
            self.required_config_keys = resnet.required_config_keys
            self.data_transforms = resnet.data_transforms
        elif model_name in vgg.model_names:
            self.model = vgg.get_model(model_name)
            self.criterion = vgg.criterion
            self.required_config_keys = vgg.required_config_keys
            self.data_transforms = vgg.data_transforms
        elif model_name in vision_transformer.model_names:
            self.model = vision_transformer.get_model(model_name)
            self.criterion = vision_transformer.criterion
            self.required_config_keys = vision_transformer.required_config_keys
            self.data_transforms = vision_transformer.data_transforms
        else:
            raise Exception("Unrecognized model name: {}".format(model_name))

    def _run_single_iteration(self, train_idx, test_idx, num_fold=0):
        self._setup_mlflow_run(num_fold)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=INIT_LR)

        self.model.to(self.device)

        self._create_dataloaders(test_idx, train_idx)

        for epoch_num in range(self.config["num_epochs"]):
            print(
                "********************************************************************************"
            )
            print(
                make_surrounded(
                    "Starting Epoch {} of {}".format(
                        epoch_num + 1, self.config["num_epochs"]
                    )
                )
            )
            print(
                "********************************************************************************"
            )
            self.run_training(epoch_num)

            validation_metrics = self.run_validation()

            print_metrics(validation_metrics, "Validation")

            self.log_mlflow_metrics(validation_metrics, "val_")
            self.validation_metrics.append(validation_metrics)

        metrics = self.run_test()
        print_metrics(metrics, "Test")

        self.log_mlflow_metrics(metrics)
        mlflow.end_run("FINISHED")

        return metrics

    def run_single_iteration(self, use_portion=1.0):
        self.data_use_portion = use_portion
        self._do_run_setup()

        iteration_start_time = datetime.now()
        self.kfold_start_time = datetime.now()

        dataset_idx = list(range(len(self.dataset)))
        train_idx, test_idx = self._split_array(dataset_idx, self.train_test_split)

        metrics = self._run_single_iteration(train_idx, test_idx, num_fold=0)
        metrics["total_runtime"] = (datetime.now() - iteration_start_time).seconds
        mean_validation_metrics = calculate_mean_metrics(self.validation_metrics)

        return {"test": metrics, "validation": mean_validation_metrics}

    def run_kfold(self, use_portion=1.0, num_folds=5):
        self.k_folds = num_folds
        self.data_use_portion = use_portion

        self._do_run_setup()

        iteration_start_time = datetime.now()

        # 20230627 - I don't remember why I added this. commenting it out for now, may need to put it back
        # this keeps OOM-ing before it hits the 5000 item print. not sure why.
        # torch.multiprocessing.set_sharing_strategy('file_system')

        print("Dataset Initial Load Complete:  {}".format(len(self.dataset)))

        kf = KFold(n_splits=self.k_folds, shuffle=True)
        for fold, (train_idx, test_idx) in enumerate(kf.split(self.dataset)):
            self.kfold_start_time = datetime.now()
            print(
                "********************************************************************************"
            )
            print(
                make_surrounded("Starting Fold {} of {}".format(fold + 1, self.k_folds))
            )
            print(
                "********************************************************************************"
            )

            metrics = self._run_single_iteration(train_idx, test_idx, num_fold=fold)
            self.test_metrics.append(metrics)

        # calculate average test and validation metrics
        mean_test_metrics = calculate_mean_metrics(self.test_metrics)
        mean_validation_metrics = calculate_mean_metrics(self.validation_metrics)

        # put total run time into the average metrics
        mean_test_metrics["total_runtime"] = (
            datetime.now() - iteration_start_time
        ).seconds

        print_metrics(mean_test_metrics, "Final (Mean)")

        return {"test": mean_test_metrics, "validation": mean_validation_metrics}

    def prepare_data(self, use_portion):
        # be lazy for now and use a pre-subsampled dataset on disk
        # will need to do better with subsampling in the future

        # can probably get rid of this default at some point, it's for backwards compat with old prepped runs
        screenshot_dataset_path = os.path.join(self.base_dir, "subsampled_screenshots")
        if "dataset_path" in self.config.keys():
            screenshot_dataset_path = self.config["dataset_path"]

        screenshot_dataset = datasets.ImageFolder(
            screenshot_dataset_path, transforms.Compose(self.data_transforms)
        )

        if use_portion < 1.0:
            print("Loading fraction of dataset: {}".format(use_portion))
            num_samples = math.ceil(use_portion * len(screenshot_dataset))
            sample_indices = range(num_samples)
            screenshot_dataset = torch.utils.data.Subset(
                screenshot_dataset, sample_indices
            )

        self.dataset = screenshot_dataset

    def run_training(self, epoch_num):
        print("Training model")
        # put the model in training mode
        self.model.train()

        # initialize
        running_loss = 0
        num_train = 0
        train_correct = 0
        num_batches = len(self.train_dataloader)

        for i, data in enumerate(self.train_dataloader, 0):
            inputs, labels = data[0].to(self.device), data[1].float().to(self.device)
            num_train += inputs.shape[0]

            for param in self.model.parameters():
                param.grad = None

            notflat_outputs = self.model(inputs)
            outputs = torch.flatten(notflat_outputs)

            loss = self.criterion(outputs, labels)
            loss.backward()
            self.optimizer.step()

            running_loss += loss.item()

            if i % self.reporting_interval == 0 and i != 0:
                ts = datetime.now().strftime(DATE_FORMAT)
                print(
                    "[{}] Epoch {}, batch {}/{}  loss: {}".format(
                        ts,
                        epoch_num + 1,
                        i,
                        num_batches,
                        running_loss / self.reporting_interval,
                    )
                )
                running_loss = 0.0
        print("Num training elements (epoch {}): {}".format(epoch_num, num_train))
        print("")

    def run_test(self):
        print("Starting test")

        all_labels = []
        all_outputs = []
        all_predictions = []
        with torch.no_grad():
            self.model.eval()

            for data in self.test_dataloader:
                inputs, labels = data[0].to(self.device), data[1].to(self.device)

                outputs = torch.flatten(self.model(inputs))

                all_labels.extend(labels.cpu())
                all_outputs.extend(outputs.cpu())

                predicted = torch.where(outputs > 0.5, 1, 0)
                all_predictions.extend(predicted.cpu())

        runtime = (datetime.now() - self.kfold_start_time).seconds

        metrics = {
            "accuracy": accuracy_score(all_labels, all_predictions),
            "precision": precision_score(
                all_labels, all_predictions, zero_division=0.0
            ),
            "recall": recall_score(all_labels, all_predictions, zero_division=0.0),
            "f1": f1_score(all_labels, all_predictions, zero_division=0.0),
            "items": len(all_labels),
            "label_sum": torch.sum(torch.stack(all_labels)),
            "output_sum": torch.sum(torch.stack(all_outputs)),
            "prediction_sum": torch.sum(torch.stack(all_predictions)),
            "labels_size": len(all_labels),
            "predictions_size": len(all_predictions),
            "runtime": runtime,
        }

        return metrics

    def run_validation(self):
        print("Starting Validation")
        correct = 0
        total = 0
        val_labels = []
        val_outputs = []
        val_predictions = []
        with torch.no_grad():
            self.model.eval()

            for data in self.val_dataloader:
                inputs, labels = data[0].to(self.device), data[1].to(self.device)
                outputs = torch.flatten(self.model(inputs))

                val_labels.extend(labels.cpu())
                val_outputs.extend(outputs.cpu())

                predicted = torch.where(outputs > 0.5, 1.0, 0.0)
                val_predictions.extend(predicted.cpu())
                total += labels.size(0)
                correct_predictions = predicted.cpu() == labels.cpu()
                correct += correct_predictions.sum().item()

        metrics = {
            "accuracy": accuracy_score(val_labels, val_predictions),
            "precision": precision_score(
                val_labels, val_predictions, zero_division=0.0
            ),
            "recall": recall_score(val_labels, val_predictions, zero_division=0.0),
            "f1": f1_score(val_labels, val_predictions, zero_division=0.0),
            "items": total,
            "label_sum": torch.sum(torch.stack(val_labels)),
            "output_sum": torch.sum(torch.stack(val_outputs)),
            "prediction_sum": torch.sum(torch.stack(val_predictions)),
            "labels_size": len(val_labels),
            "predictions_size": len(val_predictions),
        }

        self.validation_outputs.append(val_outputs)

        return metrics
