#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

from datetime import datetime
import json
import os
import pprint

import click
import mlflow
import numpy as np
import torch

from openqa_classifier import DATE_FORMAT, make_surrounded
from openqa_classifier.classifiers import (
    torch_classifier,
    huggingface_classifier,
    calculate_mean_metrics,
)

desired_mean_metric_keys = ["accuracy", "precision", "recall", "f1", "runtime", "loss"]


def run_huggingface_training(base_dir, config, output_dir, num_runs=1):
    print("")
    print("Running hugging face training {} times".format(num_runs))

    all_validation_metrics = []
    all_test_metrics = []

    # repeat_times from cli overrides config - scuffed but it works well enough for now
    actual_num_runs = num_runs

    if num_runs == 1:
        actual_num_runs = config["num_runs"]

    current_run_num = 0
    while current_run_num < actual_num_runs:
        print(
            "********************************************************************************"
        )
        print(
            make_surrounded(
                "Huggingface Trainer run {} of {}".format(
                    current_run_num + 1, actual_num_runs
                )
            )
        )
        print(
            make_surrounded(
                "Start Time: {}".format(datetime.now().strftime(DATE_FORMAT))
            )
        )
        print(make_surrounded("Model Class: {}".format(config["model_class"])))
        print(make_surrounded("Model Name: {}".format(config["model_name"])))
        print(
            "********************************************************************************"
        )
        print("")

        classifier = huggingface_classifier.OpenQAClassifierVIT(
            base_dir, config, output_dir
        )

        classifier.prepare_data(1.0)
        classifier.prep_training(iteration=current_run_num + 1)

        print("Training model")
        print("")
        val_metrics = classifier.run_training()

        validation_metrics = process_validation_metrics(val_metrics)
        train_runtime = val_metrics["train_runtime"]

        mean_validation_metrics = filter_metrics_dict(val_metrics["mean"])

        for i in range(len(validation_metrics)):
            val_metrics_filename = "val_metrics_{}-{}.json".format(current_run_num, i)
            metrics_path = os.path.join(output_dir, val_metrics_filename)
            all_validation_metrics.append(validation_metrics[i])
            with open(metrics_path, "w") as metrics_file:
                json.dump(validation_metrics[i], metrics_file)

        mean_metrics_filename = "val_metrics_mean-{}.json".format(current_run_num)
        metrics_path = os.path.join(output_dir, mean_metrics_filename)
        with open(metrics_path, "w") as metrics_file:
            json.dump(mean_validation_metrics, metrics_file)

        print("")
        print("Validation Metrics: ")
        for key in mean_validation_metrics.keys():
            print("   {}:  {}".format(key, mean_validation_metrics[key]))
        print("")
        print("Testing Model")
        print("")

        test_metrics = classifier.run_eval()

        desired_metrics = {}
        for key in test_metrics.keys():
            root_key = key[5:]
            if root_key in desired_mean_metric_keys:
                desired_metrics[root_key] = test_metrics[key]
        all_test_metrics.append(desired_metrics)

        test_metrics_filename = "test_metrics-{}.json".format(current_run_num)
        test_metrics_path = os.path.join(output_dir, test_metrics_filename)
        with open(test_metrics_path, "w") as metrics_file:
            json.dump(desired_metrics, metrics_file)

        """
        The HF trainer only reports training metrics to mlflow. Other data needs to be submitted manually
        
        WARNING: this might not work if there are multiple runs going at about the same time. mlflow docs say that
        last_active_run() will return the most recently ended run from _this_ process so it might work
        """
        mlflow_metrics = {}
        for key in desired_metrics.keys():
            mlflow_metrics["test_{}".format(key)] = desired_metrics[key]

        last_run = mlflow.last_active_run()
        last_run_id = last_run.info.run_id
        with mlflow.start_run(last_run_id) as continued_run:
            mlflow.log_metrics(mlflow_metrics)

        print("")
        print("Test Metrics:")
        for key in desired_metrics.keys():
            print("   {}:  {}".format(key, desired_metrics[key]))

        print("")

        current_run_num += 1

    mean_validation_metrics = calculate_mean_dict_metrics(all_validation_metrics)
    mean_test_metrics = calculate_mean_dict_metrics(all_test_metrics)
    mean_test_metrics["total_runtime"] = train_runtime

    return {"validation": mean_validation_metrics, "test": mean_test_metrics}


def process_validation_metrics(val_metrics):
    validation_metrics = []

    for key in val_metrics.keys():
        if key.startswith("epoch_"):
            desired_val_metrics = filter_metrics_dict(val_metrics[key])
            validation_metrics.append(desired_val_metrics)

    return validation_metrics


def calculate_mean_dict_metrics(all_metrics):
    raw_metrics = {}
    for metrics_dict in all_metrics:
        for key in metrics_dict.keys():
            if key not in raw_metrics:
                raw_metrics[key] = []
            raw_metrics[key].append(metrics_dict[key])
    mean_metrics = {}
    for key in raw_metrics.keys():
        mean_metrics[key] = np.mean(raw_metrics[key])

    return mean_metrics


def filter_metrics_dict(metrics_dict):
    desired_mean_metrics = {}
    for key in metrics_dict:
        if key in desired_mean_metric_keys:
            desired_mean_metrics[key] = metrics_dict[key]
    return desired_mean_metrics


@click.command()
@click.argument("config_path")
@click.option(
    "--base-dir", default="../openqa_data/", help="base directory for the project"
)
@click.option("--data-fraction", default=1.0, help="fraction of data to use")
@click.option(
    "--force-cpu/--no-force-cpu", default=False, help="force use of CPU over GPU"
)
@click.option(
    "--cuda-memory-snapshot-file",
    default="",
    help="file to save cuda memory snapshot into",
)
@click.option(
    "--dump-stats/--no-dump-stats", default=True, help="Dump mean metrics to file"
)
@click.option(
    "--num-folds",
    default=5,
    type=click.IntRange(1, 10),
    help="Number of folds to use for k_folds cross validation (setting to 1 only runs one iteration, "
    "no cross-validation). only works with torch classifier",
)
@click.option(
    "--num-runs",
    default=1,
    type=click.IntRange(1, 10),
    help="Number of times to run. >1 repeats identical run. only works with huggingface classifier",
)
def classifier_cli(
    config_path,
    base_dir,
    data_fraction,
    force_cpu,
    num_folds,
    dump_stats,
    cuda_memory_snapshot_file,
    num_runs,
):
    print(
        "********************************************************************************"
    )
    print(make_surrounded("OpenQA Classifier Run"))
    print(
        make_surrounded("Start Time: {}".format(datetime.now().strftime(DATE_FORMAT)))
    )
    print(make_surrounded("Config File: {}".format(config_path)))
    print(
        "********************************************************************************"
    )

    if not os.path.isfile(config_path):
        raise Exception("ERROR: Config file does not exist: {}".format(config_path))

    config = None
    with open(config_path, "r") as config_file:
        config = json.load(config_file)

    if config is None:
        raise Exception("ERROR: No config was parsed from file {}".format(config_path))

    if force_cpu:
        config["force_cpu"] = True

    output_dir = os.path.dirname(config_path)

    # old configs don't have model_name, set cnn as default
    model_name = config.get("model_name", "cnn")

    # old configs don't have trainer, set torch as default
    trainer_type = config.get("trainer_type", "torch")

    mean_metrics = {}

    if trainer_type == "huggingface":
        mean_metrics = run_huggingface_training(base_dir, config, output_dir, num_runs)
    elif trainer_type == "torch":
        classifier = torch_classifier.OpenQATorchClassifier(
            base_dir, config, output_dir
        )

        # only keep track of memory history if we're dumping a snapshot
        if cuda_memory_snapshot_file != "":
            torch.cuda.memory._record_memory_history()

        if num_folds == 1:
            mean_metrics = classifier.run_single_iteration(use_portion=data_fraction)
        else:
            mean_metrics = classifier.run_kfold(
                use_portion=data_fraction, num_folds=num_folds
            )

        if cuda_memory_snapshot_file != "":
            torch.cuda.memory._dump_snapshot("my_snapshot.pickle")

    else:
        raise Exception("Invalid trainer type used: {}".format(trainer_type))

    if dump_stats:
        dump_stats_to_file(mean_metrics, output_dir)
    else:
        pprint.pprint(mean_metrics)


def dump_stats_to_file(mean_metrics, output_dir):
    metrics = {}
    for metric_type in mean_metrics.keys():
        metrics = mean_metrics[metric_type]

        # the json serializer doesn't understand/tolerate the np.float32 that are in the metrics dict
        string_mean_metrics = {}
        for k in metrics.keys():
            string_mean_metrics[k] = str(metrics[k])

        # dump mean metrics to a json file
        mean_metrics_path = os.path.join(
            output_dir, "{}_mean_metrics.json".format(metric_type)
        )
        with open(mean_metrics_path, "w") as mean_metrics_file:
            json.dump(string_mean_metrics, mean_metrics_file)


if __name__ == "__main__":
    classifier_cli()
