#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import os
import random
import shutil


import click
import pandas as pd
from PIL import Image

supported_image_formats = [".png"]


def copy_images(image_paths, new_dir):
    if not os.path.isdir(new_dir):
        print("making new dataset dir: {}".format(new_dir))
        os.makedirs(new_dir)

    for image_path in image_paths:
        image_filename = os.path.basename(image_path)
        new_image_path = os.path.join(new_dir, image_filename)

        shutil.copyfile(image_path, new_image_path)


def balance_dataset(base_dir, source_data_path, dataset_name, not_ratio):
    """
    not_ratio - the ratio of not_rescheduled images to rescheduled images"""

    if not os.path.isdir(source_data_path):
        raise Exception(
            "source dataset path does not exist: {}".format(source_data_path)
        )

    rescheduled_dataset_dir = os.path.join(source_data_path, "rescheduled")
    not_rescheduled_dataset_dir = os.path.join(source_data_path, "not_rescheduled")

    rescheduled_images = []
    for imagefilename in os.scandir(rescheduled_dataset_dir):
        if not os.path.splitext(imagefilename.path)[1] in supported_image_formats:
            print("image file {} is not of a supported type".format(imagefilename.path))

        rescheduled_images.append(imagefilename.path)

    not_rescheduled_images = []
    for imagefilename in os.scandir(not_rescheduled_dataset_dir):
        if not os.path.splitext(imagefilename.path)[1] in supported_image_formats:
            print("image file {} is not of a supported type".format(imagefilename.path))

        not_rescheduled_images.append(imagefilename.path)

    num_not_rescheduled_images = len(rescheduled_images) * not_ratio

    selected_not_rescheduled_images = random.sample(
        not_rescheduled_images, num_not_rescheduled_images
    )

    new_rescheduled_dir = os.path.join(base_dir, dataset_name, "rescheduled")
    copy_images(rescheduled_images, new_rescheduled_dir)

    new_not_rescheduled_dir = os.path.join(base_dir, dataset_name, "not_rescheduled")
    copy_images(selected_not_rescheduled_images, new_not_rescheduled_dir)


def categorize_dataset(
    base_dir, source_data_path, dataset_name, only_use_failures=False
):
    if not os.path.isdir(source_data_path):
        raise Exception("Source data dir does not exist: {}".format(source_data_path))

    api_csv_filename = os.path.join(base_dir, "openqa_api_data.csv")
    if not os.path.isfile(api_csv_filename):
        raise Exception("api csv file does not exist: {}".format(api_csv_filename))
    api_data = pd.read_csv(api_csv_filename, index_col=0)

    dataset_path = os.path.join(base_dir, dataset_name)
    if os.path.isdir(dataset_path):
        print(
            "WARNING: dataset path already exists, may overwrite data: {}".format(
                dataset_path
            )
        )
    else:
        os.makedirs(dataset_path)

    rescheduled_path = os.path.join(dataset_path, "rescheduled")
    if os.path.isdir(rescheduled_path):
        print(
            "WARNING: rescheduled dataset dir already exists, may overwrite data: {}".format(
                rescheduled_path
            )
        )
    else:
        os.makedirs(rescheduled_path)

    not_rescheduled_path = os.path.join(dataset_path, "not_rescheduled")
    if os.path.isdir(not_rescheduled_path):
        print(
            "WARNING: not_rescheduled dataset dir already exists, may overwrite data: {}".format(
                not_rescheduled_path
            )
        )
    else:
        os.makedirs(not_rescheduled_path)

    # move images to 'rescheduled' and 'not_rescheduled' directories
    for data_row in api_data.iterrows():
        job_id = data_row[1]["job_id"]
        module_name = data_row[1]["module"]

        module_screenshot_filename = "{}.{}.composite.png".format(job_id, module_name)
        old_filename = os.path.join(source_data_path, module_screenshot_filename)

        # the api data doesn't match if there were download issues, this is a quick way to filter out invalid files
        if not os.path.isfile(old_filename):
            continue

        module_rescheduled = True if data_row[1]["rescheduled_as"] != -1 else False
        module_failed = (
            True if data_row[1]["module_result"] in ["failed", "softfailed"] else False
        )

        new_filename = None

        if module_rescheduled and module_failed:
            new_filename = os.path.join(rescheduled_path, module_screenshot_filename)
        else:
            if only_use_failures:
                if module_failed:
                    new_filename = os.path.join(
                        not_rescheduled_path, module_screenshot_filename
                    )
            else:
                new_filename = os.path.join(
                    not_rescheduled_path, module_screenshot_filename
                )

        if new_filename is not None:
            shutil.copyfile(old_filename, new_filename)


def scale_for_pretrained(base_dir, source_dataset_name, dataset_name):
    source_dataset_path = os.path.join(base_dir, source_dataset_name)
    if not os.path.isdir(source_dataset_path):
        raise Exception(
            "Source dataset dir does not exist: {}".format(source_dataset_path)
        )

    dataset_path = os.path.join(base_dir, dataset_name)
    if os.path.isdir(dataset_path):
        raise Exception(
            "ERROR: dataset output directory already exists: {}".format(dataset_path)
        )
    else:
        print("Creating dataset dir: ")
        os.mkdir(dataset_path)

    categories = []
    for entry in os.scandir(source_dataset_path):
        if entry.is_dir():
            categories.append(entry.name)

    for category in categories:
        print("Processing images in category: {}".format(category))
        source_dir = os.path.join(source_dataset_path, category)
        dest_dir = os.path.join(dataset_path, category)
        if not os.path.isdir(dest_dir):
            os.mkdir(dest_dir)

        for entry in os.scandir(source_dir):
            source_filename = entry.name
            source_image_path = os.path.join(source_dir, source_filename)
            dest_image_path = os.path.join(dest_dir, source_filename)

            source_image = Image.open(source_image_path)
            dest_image = source_image.resize((224, 224))

            dest_image.save(dest_image_path)


@click.command()
@click.argument("source_data_dir")
@click.argument("dataset_name")
@click.option(
    "--base-dir", default="../openqa_data", help="base directory for the project"
)
@click.option(
    "--only-failures/--no-only-failures",
    default=False,
    help="Only include failed jobs in dataset",
)
def categorize(base_dir, source_data_dir, dataset_name, only_failures):
    """Takes a raw dataset of composite screenshots in SOURCE_DATA_DIR directory and API data in BASE_DIR and creates a
    dataset named DATASET_NAME consisting of "rescheduled" and "not_rescheduled" screenshots in class-named directories
    """
    categorize_dataset(
        os.path.abspath(base_dir),
        os.path.abspath(source_data_dir),
        dataset_name,
        only_failures,
    )


@click.command()
@click.argument("source_data_dir")
@click.argument("dataset_name")
@click.option(
    "--base-dir", default="../openqa_data/", help="base directory for the project"
)
@click.option(
    "--ratio",
    default=4,
    type=int,
    help="number of not_rescheduled images for every reschedule image",
)
def balance(base_dir, source_data_dir, dataset_name, ratio):
    balance_dataset(base_dir, source_data_dir, dataset_name, ratio)


@click.command()
@click.argument("source_dataset_name")
@click.argument("dataset_name")
@click.option(
    "--base-dir",
    default="../openqa_data/",
    type=str,
    help="base directory for the project",
)
def prepare_pretrained(base_dir, source_dataset_name, dataset_name):
    """Takes an existing dataset of images in directories named for their categories. Scales them to 256x256 for
    compatiblity with pretrained models"""

    scale_for_pretrained(base_dir, source_dataset_name, dataset_name)


@click.group
def cli():
    pass


cli.add_command(categorize)
cli.add_command(balance)
cli.add_command(prepare_pretrained)

if __name__ == "__main__":
    cli()
