#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms

model_names = ["cnn_3_layer"]
required_config_keys = [
    "iteration_id",
    "cnn1_output_channels",
    "cnn2_output_channels",
    "cnn3_output_channels",
    "layer_1_size",
    "kernel_size",
    "max_pool_kernel",
    "max_pool_stride",
    "batch_size",
    "cnn1_input_channels",
    "layer_end_size",
    "cnn_stride",
    "num_epochs",
    "image_height",
    "image_width",
    "force_cpu",
    "base_dir",
    "random_seed",
    "mlflow_experiment_name",
    "dataset_path",
    "iteration_id",
]

criterion = nn.BCELoss()
data_transforms = [transforms.ToTensor()]


def calculate_cnn_outsize(insize, dilation, padding, kernel_size, stride):
    outsize = (insize + 2 * padding - dilation * (kernel_size - 1) - 1) / stride + 1
    return math.floor(outsize)


class OpenqaRescueduleClassifier(nn.Module):
    def __init__(self, config):
        super().__init__()

        self.debug = False
        self.config = config

        # assume that dilation=1 and padding=0 for all cnn and maxpool layers
        cnn_kernel_size = (config["kernel_size"], config["kernel_size"])
        # assume that stride is an int -> equal in all dimensions
        cnn_stride = (config["cnn_stride"], config["cnn_stride"])
        maxpool_kernel = (config["max_pool_kernel"], config["max_pool_kernel"])
        maxpool_stride = (config["max_pool_stride"], config["max_pool_stride"])

        # let's start with a 3 layer cnn, relu activation, maxpooling after every cnn layer, mlp penultimate layer
        # 2 length (2 class) linear final layer
        self.cnn1 = nn.Conv2d(
            self.config["cnn1_input_channels"],
            self.config["cnn1_output_channels"],
            kernel_size=self.config["kernel_size"],
        )
        self.maxpool1 = nn.MaxPool2d(
            kernel_size=self.config["max_pool_kernel"],
            stride=self.config["max_pool_stride"],
        )

        cnn1_height_outsize = calculate_cnn_outsize(
            config["image_height"], 1, 0, cnn_kernel_size[0], cnn_stride[0]
        )
        maxpool1_height_outsize = calculate_cnn_outsize(
            cnn1_height_outsize, 1, 0, maxpool_kernel[0], maxpool_stride[0]
        )
        cnn1_width_outsize = calculate_cnn_outsize(
            config["image_width"], 1, 0, cnn_kernel_size[1], cnn_stride[1]
        )
        maxpool1_width_outsize = calculate_cnn_outsize(
            cnn1_width_outsize, 1, 0, maxpool_kernel[1], maxpool_stride[1]
        )

        self.cnn2 = nn.Conv2d(
            self.config["cnn1_output_channels"],
            self.config["cnn2_output_channels"],
            kernel_size=self.config["kernel_size"],
        )
        self.maxpool2 = nn.MaxPool2d(
            kernel_size=self.config["max_pool_kernel"],
            stride=self.config["max_pool_stride"],
        )

        cnn2_height_outsize = calculate_cnn_outsize(
            maxpool1_height_outsize, 1, 0, cnn_kernel_size[0], cnn_stride[0]
        )
        maxpool2_height_outsize = calculate_cnn_outsize(
            cnn2_height_outsize, 1, 0, maxpool_kernel[0], maxpool_stride[0]
        )
        cnn2_width_outsize = calculate_cnn_outsize(
            maxpool1_width_outsize, 1, 0, cnn_kernel_size[1], cnn_stride[1]
        )
        maxpool2_width_outsize = calculate_cnn_outsize(
            cnn2_width_outsize, 1, 0, maxpool_kernel[1], maxpool_stride[1]
        )

        self.cnn3 = nn.Conv2d(
            self.config["cnn2_output_channels"],
            self.config["cnn3_output_channels"],
            kernel_size=self.config["kernel_size"],
        )
        self.maxpool3 = nn.MaxPool2d(
            kernel_size=self.config["max_pool_kernel"],
            stride=self.config["max_pool_stride"],
        )

        cnn3_height_outsize = calculate_cnn_outsize(
            maxpool2_height_outsize, 1, 0, cnn_kernel_size[0], cnn_stride[0]
        )
        maxpool3_height_outsize = calculate_cnn_outsize(
            cnn3_height_outsize, 1, 0, maxpool_kernel[0], maxpool_stride[0]
        )
        cnn3_width_outsize = calculate_cnn_outsize(
            maxpool2_width_outsize, 1, 0, cnn_kernel_size[1], cnn_stride[1]
        )
        maxpool3_width_outsize = calculate_cnn_outsize(
            cnn3_width_outsize, 1, 0, maxpool_kernel[1], maxpool_stride[1]
        )

        flattened_maxpool_outsize = (
            self.config["cnn3_output_channels"]
            * maxpool3_height_outsize
            * maxpool3_width_outsize
        )

        self.fc1 = nn.Linear(flattened_maxpool_outsize, self.config["layer_1_size"])
        self.fc2 = nn.Linear(self.config["layer_1_size"], self.config["layer_end_size"])

        self.output_layer = nn.Sigmoid()

    def forward(self, x):
        if self.debug:
            print("initial x: {}".format(x.shape))
        x = F.relu(self.cnn1(x))
        if self.debug:
            print("x post relu(cnn1): {}".format(x.shape))
        x = self.maxpool1(x)
        if self.debug:
            print("x post maxpool1: {}".format(x.shape))
        x = F.relu(self.cnn2(x))
        if self.debug:
            print("x post relu(cnn2): {}".format(x.shape))
        x = self.maxpool2(x)
        if self.debug:
            print("x post maxpool2: {}".format(x.shape))
        x = F.relu(self.cnn3(x))
        if self.debug:
            print("x post relu(cnn3): {}".format(x.shape))
        x = self.maxpool3(x)
        if self.debug:
            print("x post maxpool2: {}".format(x.shape))

        x = torch.flatten(x, start_dim=1)
        if self.debug:
            print("x post flatten: {}".format(x.shape))
        x = F.relu(self.fc1(x))
        if self.debug:
            print("x post final relu: {}".format(x.shape))
        x = self.fc2(x)

        output = self.output_layer(x)

        return output
