#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import argparse
import html
import io
import json
import math
import os
import sys

import click

# from pprint import pprint
import re
from urllib.error import HTTPError

# from bs4 import BeautifulSoup
# from lxml import etree
from PIL import Image

# import pytesseract
import requests

import numpy as np

from urllib.request import urlretrieve

from fedora_messaging import config
from fedora_messaging.exceptions import HaltConsumer


BASE_URL = "https://openqa.fedoraproject.org"
# BASE_DIR = '/home/tflink/code/testresult-gather/data'
BASE_DIR = "/mnt/usbstick/openqa_data"


class ProcessOpenQAMessage:
    """
    A fedora-messaging consumer that extracts the job number from openqa job done messages and
    downloads the non-video and non-image data associated with that job
    """

    def __init__(self):
        # check for the base dir
        if not os.path.exists(BASE_DIR):
            raise HaltConsumer(
                "The base dir {} does not exist, terminating".format(BASE_DIR)
            )

        self.ignorefile_re = re.compile(r".*\.(iso)|(ogv)")
        self.validlink_re = re.compile(r"/tests/([0-9]+)/file/([-0-9a-zA-Z.=/_]+[^?]+)")

    def __call__(self, message):
        """
        Invoked when an openqa message is recieved
        """

        self.process_message(message.body)

    def process_message(self, message_body):
        job_id = message_body["id"]

        gather = GatherOpenQAData(job_id, BASE_DIR)
        gather.download_and_process()

        output_dir = "{}/{}/".format(BASE_DIR, job_id)

        # dump job data to json file
        job_jsonfilename = "{}message-{}.json".format(output_dir, job_id)
        if not os.path.isfile(job_jsonfilename):
            with open(job_jsonfilename, "w") as job_jsonfile:
                json.dump(message_body, job_jsonfile)


class GatherOpenQAData:
    def __init__(self, job_id, base_dir, base_url="https://openqa.fedoraproject.org"):
        self.job_id = job_id
        self.base_dir = base_dir
        self.base_url = base_url
        self.output_dir = "{}/{}/".format(self.base_dir, job_id)
        self.max_retries = 2

        # check for the base dir
        if not os.path.exists(self.base_dir):
            raise HaltConsumer(
                "The base dir {} does not exist, terminating".format(self.base_dir)
            )

        self.isofile_re = re.compile(r".*\.(iso)")
        self.isovideofile_re = re.compile(r".*\.(iso)|(ogv)")
        self.validlink_re = re.compile(r"/tests/([0-9]+)/file/([-0-9a-zA-Z.=/_]+[^?]+)")

    def _do_request(self, url, test_json=False):
        num_retries = 0

        while num_retries < self.max_retries:
            try:
                r = requests.get(url)
                if test_json:
                    data = r.json()

                return r
            except (requests.HTTPError, requests.ConnectionError):
                print("Error during request to {}".format(url))
                num_retries += 1
            except requests.exceptions.JSONDecodeError:
                print("Error decoding json from {}".format(url))
                num_retries += 1

        print(
            "ERROR: exceeded {} retries for request to {}".format(self.max_retries, url)
        )

    def _download_file(self, url, dest):
        num_retries = 0

        while num_retries < self.max_retries:
            try:
                urlretrieve(url, dest)
                return
            except HTTPError:
                print("Error during download of {}".format(url))
                num_retries += 1

        print(
            "ERROR: exceeded {} retries during download of {}".format(
                self.max_retries, url
            )
        )

    def _check_output_dirs(self):
        if not os.path.isdir(self.output_dir):
            os.mkdir(self.output_dir)

    def download_logs_files(self, get_videos=True):
        from bs4 import BeautifulSoup
        from lxml import etree

        print("Downloading logs and files from job {}".format(self.job_id))

        job_url = "{}/tests/{}/downloads_ajax".format(self.base_url, self.job_id)
        self._check_output_dirs()

        # get raw html about job in question
        r = self._do_request(job_url)
        job_text = r.text

        soup = BeautifulSoup(io.StringIO(job_text), "lxml")

        # find all the links - for now, they're all in <ul>
        lists = soup.find_all("ul")
        all_links = []
        for l in lists:
            all_links.extend(l.find_all("a"))

        # download all the links which aren't videos or isos
        for link in all_links:
            link_match = self.validlink_re.search(link.attrs["href"])

            if link_match:
                if get_videos:
                    bigfile_match = self.isofile_re.search(link.attrs["href"])
                else:
                    bigfile_match = self.isovideofile_re.search(link.attrs["href"])

                if bigfile_match:
                    continue

                filename = link_match.group(2)
                output_file = "{}{}".format(self.output_dir, filename)
                file_url = "{}{}".format(self.base_url, link.attrs["href"])

                if not os.path.isfile(output_file):
                    # print("dowloading link: {} to {}".format(file_url, output_file))
                    self._download_file(file_url, output_file)

    def download_screenshots(self):
        print("Downloading screenshots for job {}".format(self.job_id))

        job_details_url = "{}/tests/{}/details_ajax".format(self.base_url, self.job_id)
        r = self._do_request(job_details_url, test_json=True)
        job_details = r.json()

        screenshot_filenames = []

        # check to see if the job is incomplete
        if not "modules" in job_details.keys():
            print("No screenshots detected for job {}".format(self.job_id))
            return

        for module in job_details["modules"]:
            screenshots = []
            for detail in module["details"]:
                if "screenshot" in detail.keys():
                    screenshots.append(detail["screenshot"])
            screenshot_filenames.extend(screenshots)

        # download all the screenshots
        for screenshot_name in screenshot_filenames:
            screenshot_url = "{}/tests/{}/images/{}".format(
                self.base_url, self.job_id, screenshot_name
            )
            screenshot_output_dir = "{}screenshots/".format(self.output_dir)
            output_file = "{}{}".format(screenshot_output_dir, screenshot_name)

            if not os.path.isdir(screenshot_output_dir):
                os.makedirs(screenshot_output_dir)
            if not os.path.isfile(output_file):
                # print("dowloading link: {} to {}".format(screenshot_url, output_file))
                self._download_file(screenshot_url, output_file)
            else:
                pass
                # print("file {} already exists at {}".format(screenshot_url, screenshot_output_dir))

    def ocr_screenshots(self):
        import pytesseract

        print("Running OCR on screenshots from job {}".format(self.job_id))

        self._check_output_dirs()
        screenshot_dir = "{}screenshots".format(self.output_dir)

        # assume that all the screenshots are .png
        with os.scandir(screenshot_dir) as filescan:
            for entry in filescan:
                if entry.is_file():
                    filename = "{}/{}".format(screenshot_dir, entry.name)
                    if os.path.basename(filename).endswith(".png"):
                        text_output_filename = "{}.txt".format(filename)

                        if os.path.isfile(text_output_filename):
                            # print("ocr file already exists for: {}".format(filename))
                            pass
                        else:
                            # print("running OCR on {}".format(filename))
                            text = pytesseract.image_to_string(Image.open(filename))

                            with open(text_output_filename, "w") as text_output_file:
                                text_output_file.write(text)

    def download_api_data(self):
        print("Downloading API data for job {}".format(self.job_id))

        job_api_url = "{}/api/v1/jobs?ids={}".format(self.base_url, self.job_id)
        api_data_filename = "{}{}_apidata.json".format(self.output_dir, self.job_id)
        self._check_output_dirs()

        if os.path.isfile(api_data_filename):
            # print("openqa api data file already exists, skipping: {}".format(api_data_filename))
            return

        r = self._do_request(job_api_url, test_json=True)
        job_api_json_data = r.json()

        with open(api_data_filename, "w") as api_data_file:
            api_data_file.write(json.dumps(job_api_json_data))

    def _find_screenshots(self, screenshot_dir, file_extension=".png"):
        screenshots = {}
        with os.scandir(screenshot_dir) as screenshotscan:
            for screenshot_entry in screenshotscan:
                if screenshot_entry.is_file():
                    if screenshot_entry.name.endswith(file_extension):
                        # ew, this is complex - openqa names screenshots as [testname]-[number]
                        # and I want to group by the test name
                        testname = "-".join(screenshot_entry.name.split("-")[:-1])

                        if len(testname) == 0:
                            continue

                        if not testname in screenshots.keys():
                            screenshots[testname] = [screenshot_entry.name]
                        else:
                            screenshots[testname].append(screenshot_entry.name)

        return screenshots

    def make_composite_screenshots(self, end_scale=1):
        print("Making composite screenshots for job {}".format(self.job_id))

        if not end_scale.is_integer():
            raise Exception(
                "composite scaling factor must be an integer. Got {}".format(end_scale)
            )

        screenshot_dir = os.path.join(self.output_dir, "screenshots")
        screenshots = self._find_screenshots(screenshot_dir)

        screenshot_output_dir = os.path.join(screenshot_dir, "composite")
        if not os.path.isdir(screenshot_output_dir):
            os.makedirs(screenshot_output_dir)
        for category in screenshots.keys():
            num_screenshots = len(screenshots[category])

            # assume all screenshots are the same size, sample one to find the exact size
            sample_image_path = os.path.join(screenshot_dir, screenshots[category][0])
            sample_image = Image.open(sample_image_path)
            image_width, image_height = sample_image.size

            # find the number of screenshots on each side of the composite image
            composite_len = math.ceil(math.sqrt(num_screenshots))

            composite_width = composite_len * image_width
            composite_height = composite_len * image_height
            composite_image_array = np.ones(
                shape=(composite_width, composite_height), dtype="uint8"
            )
            composite_image = Image.new(
                "RGB", (composite_width, composite_height), color=(255, 255, 255)
            )

            column_i = 0
            row_i = 0

            for png_file in screenshots[category]:
                image_filename = os.path.join(screenshot_dir, png_file)
                pil_image = Image.open(image_filename)
                cpil_image = pil_image.convert("RGB")

                start_width = column_i * image_width
                start_height = row_i * image_height

                composite_image.paste(cpil_image, (start_width, start_height))

                # handle munging of the i variables
                if column_i >= composite_len:
                    row_i += 1
                    column_i = 0
                else:
                    column_i += 1

            resized_width = int(image_width * end_scale)
            resized_height = int(image_height * end_scale)
            resized_composite = composite_image.resize((resized_width, resized_height))
            composite_filename = os.path.join(
                screenshot_output_dir,
                "{}.{}.composite.{}x.png".format(self.job_id, category, end_scale),
            )
            resized_composite.save(composite_filename)

    def download_and_process(self):
        self.download_api_data()
        self.download_logs_files()
        self.download_screenshots()
        self.make_composite_screenshots()
        self.ocr_screenshots()
        print()


@click.command()
@click.argument("jobid")
@click.option("-b", "--basedir", default=".", help="Parent directory to all job data")
@click.option(
    "--with-screenshots/--skip-screenshots",
    default=True,
    help="download screenshots from openqa",
)
@click.option(
    "--with-composite/--skip-composite",
    default=True,
    help="create composite screenshots",
)
@click.option("--with-logs/--skip-logs", default=True, help="download logs from openqa")
@click.option("--with-ocr/--skip-ocr", default=True, help="run OCR on screenshots")
@click.option(
    "--with-api/--skip-api", default=True, help="download api data from openqa"
)
@click.option(
    "--fast-skip/--no-fast-skip",
    default=False,
    help="skip all processing if target directory exists",
)
@click.option(
    "--composite-scale",
    default=1.0,
    type=float,
    help="multiple of original resolution to use for composite screenshots",
)
def cli(
    jobid,
    basedir,
    with_screenshots,
    with_composite,
    with_logs,
    with_ocr,
    with_api,
    fast_skip,
    composite_scale,
):
    gather = GatherOpenQAData(jobid, basedir)

    if fast_skip and os.path.isdir(gather.output_dir):
        print("directory {} exists, exiting".format(gather.output_dir))
        sys.exit(0)
    if with_api:
        gather.download_api_data()
    if with_logs:
        gather.download_logs_files()
    if with_screenshots:
        gather.download_screenshots()
    if with_composite:
        gather.make_composite_screenshots(composite_scale)
    if with_ocr:
        gather.ocr_screenshots()


if __name__ == "__main__":
    cli()
