#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import json
import os

import click
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import subprocess


class RunData(object):
    def __init__(self, run_directory, remove_zero_runs=True):
        self.run_directory = run_directory
        self.output_dir = os.path.join(self.run_directory, "results")
        self.chart_dir = os.path.join(self.output_dir, "charts")
        self.data_dir = os.path.join(self.output_dir, "data")
        self.remove_zero_runs = remove_zero_runs

        self.val_mean_metrics = []
        self.test_mean_metrics = []
        self.combined_metrics = []

        self.configurations = []
        self.invalid_iterations = []
        self.pending_iterations = []
        self.bad_iterations = []
        self.bad_configurations = []
        self.cuda_oom_iterations = []
        self.cuda_oom_configurations = []

        self.time_stats = None

        self.iteration_config_df = None
        self.test_metrics_df = None
        self.val_metrics_df = None
        self.test_val_metrics_df = None

    def _load_rundir_data(self):
        for entry in os.scandir(self.run_directory):
            if not entry.is_dir():
                continue

            if entry.name == "results":
                continue

            validation_metrics_path = os.path.join(
                entry.path, "validation_mean_metrics.json"
            )
            test_metrics_path = os.path.join(entry.path, "test_mean_metrics.json")
            console_output_path = os.path.join(entry.path, "console_output")
            iteration_config_path = os.path.join(entry.path, "config.json")

            iteration_config = None
            if not os.path.isfile(iteration_config_path):
                self.invalid_iterations.append(entry.name)
                continue

            with open(iteration_config_path, "r") as configfile:
                iteration_config = json.load(configfile)

            if not os.path.isfile(console_output_path):
                self.pending_iterations.append(entry.name)
                continue

            if not os.path.isfile(validation_metrics_path) or not os.path.isfile(
                test_metrics_path
            ):
                self.bad_iterations.append(entry.name)
                self.bad_configurations.append(iteration_config)

                is_cuda_oom = subprocess.call(
                    [
                        "grep",
                        "torch.cuda.OutOfMemoryError: CUDA out of memory.",
                        console_output_path,
                    ],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )

                failure_reason = None
                if is_cuda_oom == 0:
                    self.cuda_oom_configurations.append(iteration_config)
                    self.cuda_oom_iterations.append(entry.name)

                continue

            # if we've gotten this far, there is a file for test and validation metrics so the run is valid
            self.configurations.append(iteration_config)

            validation_metrics = {}
            with open(validation_metrics_path, "r") as validation_metrics_file:
                raw_validation_metrics = json.load(validation_metrics_file)
                for key in raw_validation_metrics.keys():
                    validation_metrics[key] = float(raw_validation_metrics[key])
                validation_metrics["iteration_id"] = iteration_config["iteration_id"]

            test_metrics = {}
            with open(test_metrics_path, "r") as test_metrics_file:
                raw_test_metrics = json.load(test_metrics_file)
                for key in raw_test_metrics.keys():
                    test_metrics[key] = float(raw_test_metrics[key])
                test_metrics["iteration_id"] = iteration_config["iteration_id"]
                validation_metrics["total_runtime"] = test_metrics["total_runtime"]

            combined_metrics = {"total_runtime": test_metrics["total_runtime"]}
            combined_keys = ["accuracy", "precision", "recall", "f1"]

            for key in combined_keys:
                val_key = "val_{}".format(key)
                combined_metrics[val_key] = validation_metrics[key]

                test_key = "test_{}".format(key)
                combined_metrics[test_key] = test_metrics[key]

            combined_metrics["iteration_id"] = iteration_config["iteration_id"]

            self.combined_metrics.append(combined_metrics)
            self.val_mean_metrics.append(validation_metrics)
            self.test_mean_metrics.append(test_metrics)

    def _create_dataframes(self):
        if (
            self.iteration_config_df is None
            or self.test_mean_metrics is None
            or self.val_mean_metrics is None
            or self.combined_metrics is None
        ):
            self._load_rundir_data()

        self.iteration_config_df = pd.DataFrame(self.configurations).set_index(
            "iteration_id"
        )
        test_metrics_df = (
            pd.DataFrame(self.test_mean_metrics)
            .set_index("iteration_id")
            .join(self.iteration_config_df)
        )
        val_metrics_df = (
            pd.DataFrame(self.val_mean_metrics)
            .set_index("iteration_id")
            .join(self.iteration_config_df)
        )
        test_val_df = (
            pd.DataFrame(self.combined_metrics)
            .set_index("iteration_id")
            .join(self.iteration_config_df)
        )

        if self.remove_zero_runs:
            test_metrics_df = test_metrics_df.loc[test_metrics_df["precision"] > 0]
            val_metrics_df = val_metrics_df.loc[val_metrics_df["precision"] > 0]
            test_val_df = test_val_df.loc[test_val_df["val_precision"] > 0]

        self.test_metrics_df = test_metrics_df
        self.val_metrics_df = val_metrics_df
        self.test_val_metrics_df = test_val_df

    def _save_chart_to_disk(self, figure):
        pass

    def _generate_2d_scatter_colormap(
        self, datasource, title, x, y, c, colormap="viridis"
    ):
        fig, ax = plt.subplots()
        if datasource == "test":
            self.test_metrics_df.plot.scatter(
                x=x, y=y, c=c, colormap=colormap, ax=ax, title=title, alpha=0.5
            )
        elif datasource == "val":
            self.val_metrics_df.plot.scatter(
                x=x, y=y, c=c, colormap=colormap, ax=ax, title=title, alpha=0.5
            )

        chart_filename = "{}_{}-{}-{}_scatter.png".format(
            datasource, x.replace("_", ""), y.replace("_", ""), c.replace("_", "")
        )
        chart_path = os.path.join(self.chart_dir, chart_filename)
        plt.savefig(chart_path)

    def create_charts(self):
        if self.test_val_metrics_df is None:
            self._create_dataframes()

        if not os.path.isdir(self.chart_dir):
            print(
                "WARNING: chart dir does not exist, creating: {}".format(self.chart_dir)
            )
            os.makedirs(self.chart_dir)

        # accuracy, precision, color=total_runtime
        self._generate_2d_scatter_colormap(
            "test",
            title="Test Accuracy vs. Precision",
            x="accuracy",
            y="precision",
            c="total_runtime",
        )
        self._generate_2d_scatter_colormap(
            "val",
            title="Validation Accuracy vs. Precision",
            x="accuracy",
            y="precision",
            c="total_runtime",
        )

        # precision, recall, color=f1
        self._generate_2d_scatter_colormap(
            "test", title="Test Precision vs. Recall", x="precision", y="recall", c="f1"
        )
        self._generate_2d_scatter_colormap(
            "val",
            title="Validation Precision vs. Recall",
            x="precision",
            y="recall",
            c="f1",
        )

    def dump_csv(self):
        if (
            self.test_metrics_df is None
            or self.val_metrics_df is None
            or self.test_val_metrics_df is None
        ):
            self._create_dataframes()

        if not os.path.isdir(self.data_dir):
            print(
                "WARNING: data dir does not exist, creating: {}".format(self.data_dir)
            )
            os.makedirs(self.data_dir)

        test_metrics_path = os.path.join(self.data_dir, "test_mean_metrics.csv")
        self.test_metrics_df.to_csv(test_metrics_path)

        val_metrics_path = os.path.join(self.data_dir, "val_mean_metrics.csv")
        self.val_metrics_df.to_csv(val_metrics_path)

        test_val_metrics_path = os.path.join(self.data_dir, "test_val_mean_metrics.csv")
        self.test_val_metrics_df.to_csv(test_val_metrics_path)

    def summarize_models_used(self):
        """This is intended to make a list of the huggingface models tested and won't work on runs that used the
        torch classifier"""

        if self.test_val_metrics_df is None:
            self._create_dataframes()

        # for ease of use later, reject runs that aren't _only_ huggingface
        trainers_used = self.test_val_metrics_df.trainer_type.unique()
        if len(trainers_used) != 1 or trainers_used[0] != "huggingface":
            raise Exception("Model summary only works with huggingface training runs")

        models_used = {}
        for index, data in self.test_val_metrics_df.iterrows():
            model_class = data["model_class"]
            model_name = data["model_name"]

            if model_class not in models_used:
                models_used[model_class] = set()

            models_used[model_class].add(model_name)

        print("Models used summary:")

        for model_class in models_used:
            print("")
            print(model_class)
            models = sorted(list(models_used[model_class]))
            for model in models:
                print("  {}".format(model))

    def summarize_run(self):
        if self.time_stats is None:
            self._calculate_time_stats()

        run_name = self.configurations[0]["mlflow_experiment_name"]

        num_runs = len(self.test_val_metrics_df)

        print("Run Summary:")
        print("  name: {}".format(run_name))
        print("  successful runs:     {}".format(num_runs))
        print("  pending runs:        {}".format(len(self.pending_iterations)))
        print("  bad runs:            {}".format(len(self.bad_iterations)))
        print("  cuda oom runs:       {}".format(len(self.cuda_oom_configurations)))
        print("  mean runtime (s):    {}".format(self.time_stats["mean_runtime_secs"]))
        print(
            "  mean runtime (m):    {}".format(
                self.time_stats["mean_runtime_secs"] / 60
            )
        )
        print(
            "  median runtime (s) : {}".format(self.time_stats["median_runtime_secs"])
        )
        print(
            "  median runtime (m) : {}".format(
                self.time_stats["median_runtime_secs"] / 60
            )
        )
        print(
            "  total runtime: {} day(s), {} hour(s), {} minute(s)".format(
                self.time_stats["runtime_days"],
                self.time_stats["runtime_hours"],
                self.time_stats["runtime_mins"],
            )
        )
        print("    - seconds: {}".format(self.time_stats["total_runtime_secs"]))
        print("    - minutes: {}".format(self.time_stats["total_runtime_mins"]))
        print("    - hours:   {}".format(self.time_stats["total_runtime_hours"]))

    def _calculate_time_stats(self):
        if self.test_val_metrics_df is None:
            self._create_dataframes()

        trainer = self.configurations[0]["trainer_type"]

        self.time_stats = {}

        # want num_runs, total time
        self.time_stats["total_runtime_secs"] = np.sum(
            self.test_val_metrics_df["total_runtime"]
        )
        self.time_stats["total_runtime_mins"] = np.sum(
            self.test_val_metrics_df["total_runtime"] / 60
        )
        self.time_stats["total_runtime_hours"] = np.sum(
            self.test_val_metrics_df["total_runtime"] / (60 * 60)
        )
        if trainer == "huggingface":
            print("WARNING: huggingface trainer detected, doubling runtime")
            self.time_stats["total_runtime_secs"] *= 2
        self.time_stats["mean_runtime_secs"] = np.mean(
            self.test_val_metrics_df["total_runtime"]
        )
        self.time_stats["mean_runtime_mins"] = self.time_stats["mean_runtime_secs"] / 60
        self.time_stats["median_runtime_secs"] = np.median(
            self.test_val_metrics_df["total_runtime"]
        )
        self.time_stats["median_runtime_mins"] = (
            self.time_stats["median_runtime_secs"] / 60
        )
        self.time_stats["runtime_hours"] = self.time_stats["total_runtime_secs"] / (
            60 * 60
        )
        self.time_stats["runtime_days"] = int(
            self.time_stats["total_runtime_secs"] // (60 * 60 * 24)
        )
        self.time_stats["runtime_hours"] = int(
            (
                self.time_stats["total_runtime_secs"]
                - self.time_stats["runtime_days"] * 60 * 60 * 24
            )
            // (60 * 60)
        )
        self.time_stats["runtime_mins"] = int(
            (
                self.time_stats["total_runtime_secs"]
                - self.time_stats["runtime_days"] * 60 * 60 * 24
                - self.time_stats["runtime_hours"] * 60 * 60
            )
            // 60
        )

    def get_run_status(self, num_machines=1, show_bad=False, show_pending=False):
        if self.time_stats is None:
            self._calculate_time_stats()

        num_runs = len(self.test_val_metrics_df)
        run_name = self.configurations[0]["mlflow_experiment_name"]

        run_rate = self.time_stats["total_runtime_secs"] / num_runs
        num_pending_runs = len(self.pending_iterations)

        est_remaining_secs = run_rate * num_pending_runs
        est_remaining_hours = est_remaining_secs / (60 * 60)

        scaled_remaining_hours = est_remaining_hours / num_machines

        print("Run Status:")
        print("  name: {}".format(run_name))
        print("  successful runs:     {}".format(num_runs))
        print("  pending runs:        {}".format(len(self.pending_iterations)))
        print("  bad runs:            {}".format(len(self.bad_iterations)))
        print("  cuda oom runs:       {}".format(len(self.cuda_oom_configurations)))
        print("  seconds/run:         {}".format(run_rate))
        print("  est. hours left:     {}".format(est_remaining_hours))
        if num_machines > 1:
            print(
                "    with {} machines:  {}".format(num_machines, scaled_remaining_hours)
            )

        if show_pending:
            print("")
            print("pending runs:")
            for iteration in self.pending_iterations:
                print("  {}".format(iteration))

        if show_bad:
            print("")
            print("Bad Runs:")
            for iteration in self.bad_iterations:
                cuda_string = ""
                if iteration in self.cuda_oom_iterations:
                    cuda_string = " (cuda_oom)"

                print("  {:14}{}".format(iteration, cuda_string))


@click.command
@click.argument("run_directory")
def dump_csv(run_directory):
    run = RunData(run_directory)
    run.dump_csv()


@click.command
@click.argument("run_directory")
def charts(run_directory):
    run = RunData(run_directory)
    run.create_charts()


@click.command
@click.argument("run_directory")
def model_summary(run_directory):
    run = RunData(run_directory)
    run.summarize_models_used()


@click.command
@click.argument("run_directory")
def summary(run_directory):
    run = RunData(run_directory)
    run.summarize_run()


@click.command
@click.argument("run_directory")
@click.option("--num-machines", default=1)
@click.option("--show-bad/--no-show-bad", default=False, help="identify bad runs")
@click.option(
    "--show-pending/--no-show-pending", default=False, help="identify pending runs"
)
def run_status(run_directory, num_machines, show_bad, show_pending):
    run = RunData(run_directory, remove_zero_runs=False)
    run.get_run_status(num_machines, show_bad, show_pending)


@click.group()
def cli():
    pass


cli.add_command(dump_csv)
cli.add_command(charts)
cli.add_command(model_summary)
cli.add_command(summary)
cli.add_command(run_status)


if __name__ == "__main__":
    cli()
