#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

from abc import abstractmethod
from datetime import datetime
import itertools
import json
import os
import random
import subprocess
import time

from PIL import Image

from openqa_classifier import DATE_FORMAT, nvidia, send_email_notification


class BaseClassifierTuner(object):
    def __init__(
        self,
        base_dir,
        log_base_dir,
        dataset_name="composite_failures_1x",
        run_name_text="classifier-tuning",
    ):
        self.base_dir = os.path.abspath(base_dir)
        self.log_base_dir = os.path.abspath(log_base_dir)

        self.dataset_name = dataset_name
        self.dataset_path = os.path.join(self.base_dir, self.dataset_name)

        self.run_name_text = run_name_text
        log_dirname = "{}-{}".format(get_formatted_timestamp(), run_name_text)
        self.log_dir = os.path.join(self.log_base_dir, log_dirname)

        self.classifier_module = None
        self.configs = []
        self.base_config = {}
        self.tunables = None

        # assuming GPU0
        self.gpu_id = 0
        self.wait_secs = 10
        self.max_waits = 10

    @abstractmethod
    def _init_base_config(self, force_cpu=False):
        pass

    def _detect_screenshot_size(self):
        """detects resolution of screenshots in dataset. Assumes that all screenshots are of the same resolution"""

        screenshot_dir = os.path.join(self.dataset_path, "not_rescheduled")
        screenshot_files = os.listdir(screenshot_dir)

        example_screenshot_path = os.path.join(screenshot_dir, screenshot_files[0])

        # assume that width and height are in pixels
        with Image.open(example_screenshot_path) as image:
            info = image.info
            width = image.width
            height = image.height

        return width, height

    def _generate_runs(self):
        tunable_lists = []
        for key in self.tunables.keys():
            tunable_list = []
            for param in self.tunables[key]:
                tunable_list.append("{}:{}:{}".format(key, param, type(param)))
            tunable_lists.append(tunable_list)

        for tunable_config in itertools.product(*tunable_lists):
            decoded_config = _decode_config(tunable_config)
            decoded_config.update(self.base_config)
            self.configs.append(decoded_config)

    def _prep_logdir(self):
        if not os.path.isdir(self.log_base_dir):
            print("base log dir not found, creating {}".format(self.log_base_dir))
            os.makedirs(self.log_base_dir)

        if os.path.isdir(self.log_dir):
            print(
                "ERROR: log directory for tuning run already exists: {}".format(
                    self.log_dir
                )
            )
            raise Exception("Log directory for tuning run already exists")
        else:
            print("Creating log directory for tuning run: {}".format(self.log_dir))
            os.makedirs(self.log_dir)

    def _check_cuda_memory_and_wait(self, min_cuda_memory):
        num_waits = 0
        sufficient_memory = False
        nv_freemem = 0
        while not sufficient_memory and num_waits < self.max_waits:
            nv_freemem, nv_usedmem = nvidia.get_gpu_memory(0)
            print("GPU{} Free Memory: {} MiB".format(self.gpu_id, nv_freemem))
            # This is for one particular system
            # TODO: make this code not horrendous or remove it
            if nv_freemem < min_cuda_memory:
                print(
                    "WARNING: Insufficient free memory on GPU0, waiting {} seconds ({} / {})".format(
                        self.wait_secs, num_waits + 1, self.max_waits
                    )
                )
                time.sleep(self.wait_secs)
            else:
                sufficient_memory = True

            num_waits += 1

    def _spawn_classifier_process(self, classifier_module, iteration_logdir):
        output_path = os.path.join(iteration_logdir, "console_output")
        script_path = os.path.join(iteration_logdir, "run_classifier.sh")
        config_path = os.path.join(iteration_logdir, "config.json")

        print("Root dir for run: {}".format(iteration_logdir))
        print("Spawn process for iteration with config {}".format(config_path))
        print("Console output for run {}".format(output_path))
        # command = ["python", "-m", str(classifier_module), str(config_path), '2>&1', '|', 'tee', output_path]
        # command = ["python", "-m", str(classifier_module), str(config_path)]
        command = ["bash", script_path]
        proc = subprocess.run(
            command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        return proc

    def prep_tune(self, num_epochs, num_samples, classifier_module, force_cpu=False):
        """
        The general execution plan is:
          1. create a base log directory based on the current time
          2. for each config:
            a. ensure sufficient free memory is available w/ CUDA in torch
            b. create a new directory based on the start time for the iteration (secondarily, the config)
            c. generate a json file based on the config, put it in the new directory
            d. spawn process to run the classifier, copy all output to a file in the run directory
            e. kill child process
        """
        self._init_base_config(force_cpu)
        self._generate_runs()
        print("Generated {} configs".format(len(self.configs)))

        self._prep_logdir()

        num_configs = len(self.configs)
        for i in range(num_configs):
            config = self.configs[i]

            # create logdir for this iteration
            iteration_logdir_name = "iteration-{}".format(i)
            iteration_logdir = os.path.join(self.log_dir, iteration_logdir_name)

            if os.path.isdir(iteration_logdir):
                print(
                    "ERROR: iteration log dir already exists {}".format(
                        iteration_logdir
                    )
                )
                raise Exception("Iteration logdir already exists")
            else:
                os.makedirs(iteration_logdir)

            # dump config to file in iteration logdir
            config["iteration_id"] = i
            config_path = os.path.join(iteration_logdir, "config.json")

            with open(config_path, "w") as configfile:
                json.dump(config, configfile)

            """
            I can't get python to behave and tee output to file while showing it in real time on the console and I'm not
            spending any more time on this right now.
            
            HACKY HACKY
            
            write the command with tee to a shell script, adapt the run stuff to use the shell script
            """

            command_logfile = os.path.join(iteration_logdir, "console_output")

            command = "python -u -m {} {} 2>&1 | tee {}\n".format(
                classifier_module, config_path, command_logfile
            )
            # command = "python -u -m {} {} 2>&1 > {}\n".format(classifier_module, config_path, command_logfile)

            shell_script_path = os.path.join(iteration_logdir, "run_classifier.sh")
            with open(shell_script_path, "w") as scriptfile:
                scriptfile.write("#!/usr/bin/bash\n")
                scriptfile.write("\n")
                scriptfile.write("")
                scriptfile.write(command)
                scriptfile.write("\n")

    def run_prepped_tune(self, run_dir, classifier_module, min_cuda_memory=7000):
        """
        The idea here is to generate all the configs that we need and then execute all the generated configs. The
        restrictions in place here are:
          1. pytorch is refusing to let go of CUDA memory at times so each run will need to be in a separate child
             process which can be killed once the run is over.
          2. This is not terribly sophisticated, on purpose. The design is for a single machine running one job at a
             time. I had too many problems with a more sophisticated setup that could handle multiple machines etc.
        """

        run_start_time = datetime.now()

        if not os.path.isdir(run_dir):
            raise Exception("run_dir does not exist: {}".format(run_dir))

        num_previous_runs = 0
        # gather all config files that we're looking to run
        pending_run_configs = []
        with os.scandir(run_dir) as run_dir_scan:
            for entry in run_dir_scan:
                if entry.is_dir():
                    entry_dir = os.path.join(run_dir, entry.name)
                    mean_stats_path = os.path.join(
                        entry_dir, "validation_mean_metrics.json"
                    )
                    if os.path.isfile(mean_stats_path):
                        print(
                            "validation_mean_metrics.json file exists in directory, skipping: {}".format(
                                entry_dir
                            )
                        )
                        num_previous_runs += 1
                        continue

                    config_path = os.path.join(entry_dir, "config.json")
                    if os.path.isfile(config_path):
                        pending_run_configs.append(config_path)
                    else:
                        print(
                            "config file missing from directory, skipping: {}".format(
                                entry_dir
                            )
                        )
                        continue

        # randomize the order of the config files
        random.shuffle(pending_run_configs)

        num_failed = 0
        num_configs = num_previous_runs + len(pending_run_configs)
        for i in range(len(pending_run_configs)):
            # check for stop file
            stop_file = os.path.join(run_dir, "stop_run")
            if os.path.isfile(stop_file):
                print("WARNING: stop file found, ending tuning run")
                break

            run_config_path = pending_run_configs[i]
            with open(run_config_path, "r") as configfile:
                config = json.load(configfile)

            iteration_logdir = os.path.dirname(run_config_path)

            print(
                "==================== Run {}/{} ====================".format(
                    i + 1 + num_previous_runs, num_configs
                )
            )
            print("Start Time: {}".format(datetime.now().strftime("%Y%m%d-%H:%M.%S")))
            print("")
            # self._check_cuda_memory_and_wait(min_cuda_memory)

            console_output_path = os.path.join(iteration_logdir, "console_output")
            if os.path.isfile(console_output_path):
                print(
                    "console_output file exists, skipping iteration {}".format(
                        i + 1 + num_previous_runs
                    )
                )
                continue

            # spawn child process, copying output to file in log dir
            classifier_proc = self._spawn_classifier_process(
                classifier_module, iteration_logdir
            )

            if classifier_proc.returncode != 0:
                print(
                    "ERROR: non-zero exit code from classifier run: {}".format(
                        iteration_logdir
                    )
                )
                num_failed += 1
                continue

            test_metrics_path = os.path.join(iteration_logdir, "test_mean_metrics.json")
            if not os.path.isfile(test_metrics_path):
                print(
                    "WARNING: test_mean_metrics.json file does not exist for run: {}".format(
                        iteration_logdir
                    )
                )
                continue

            with open(test_metrics_path, "r") as metricsfile:
                test_mean_metrics = json.load(metricsfile)

            print("")
            print("Run Results:")
            print("  accuracy:  {}".format(test_mean_metrics["accuracy"]))
            print("  precision: {}".format(test_mean_metrics["precision"]))
            print("  recall:    {}".format(test_mean_metrics["recall"]))
            print("  f1:        {}".format(test_mean_metrics["f1"]))
            print("  runtime:   {}".format(test_mean_metrics["runtime"]))
            print("  total_run: {}".format(test_mean_metrics["total_runtime"]))
            print("")

            # wait between runs in an attempt to reduce the number of random-looking "out of memory" errors
            print("")
            print("Waiting {} seconds before the next run".format(WAIT_BETWEEN_RUNS))
            time.sleep(WAIT_BETWEEN_RUNS)

        run_time_dt = datetime.now() - run_start_time
        print("datetime run: {}".format(run_time_dt))
        run_time = run_time_dt.seconds
        print("total seconds: {}".format(run_time))
        run_days = run_time // 86400
        run_hours = (run_time - run_days * 86400) // 3600
        run_minutes = (run_time - run_days * 86400 - run_hours * 3600) // 60
        run_seconds = run_time - run_days * 86400 - run_hours * 3600 - run_minutes * 60

        runtime_string = "{} days, {} hours, {} minutes, {} seconds".format(
            run_days, run_hours, run_minutes, run_seconds
        )
        print("")
        print("==================== RUN COMPLETE ====================")
        print("Configs Run:    {}".format(num_configs))
        print("Configs Failed: {}".format(num_failed))
        print("Run Seconds:    {}".format(run_time))
        print("Run Time:       {}".format(runtime_string))


#        email_text = "OpenQA Classifier finished at {} in {}".format(
#            datetime.now().strftime(DATE_FORMAT), runtime_string
#        )
#        send_email_notification(email_text)


def _decode_config(config):
    new_config = {}
    for item in config:
        config_key, value, type = item.split(":")

        if type == "<class 'int'>":
            new_config[config_key] = int(value)
        elif type == "<class 'str'>":
            new_config[config_key] = value
        else:
            raise Exception("unsupported type: {}".format(type))

    return new_config


def get_formatted_timestamp():
    return datetime.now().strftime(DATE_FORMAT)


WAIT_BETWEEN_RUNS = 10
MAX_LOGFILES = 100
