#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

from datetime import datetime
import itertools

from . import BaseClassifierTuner, _decode_config
from .. import RANDOM_SEED, DATE_FORMAT

hf_model_list = {
    "swin": [
        "microsoft/swin-tiny-patch4-window7-224",
        "microsoft/swin-small-patch4-window7-224",
        "microsoft/swin-base-patch4-window7-224",
        "microsoft/swin-base-patch4-window7-224-in22k",
        "microsoft/swin-base-patch4-window12-384",
        "microsoft/swin-base-patch4-window12-384-in22k",
        "microsoft/swin-large-patch4-window7-224",
        "microsoft/swin-large-patch4-window7-224-in22k",
        "microsoft/swin-large-patch4-window12-384-in22k",
    ],
    "swinv2": [
        "microsoft/swinv2-tiny-patch4-window8-256",
        "microsoft/swinv2-tiny-patch4-window16-256",
        "microsoft/swinv2-small-patch4-window8-256",
        "microsoft/swinv2-small-patch4-window16-256",
        "microsoft/swinv2-base-patch4-window8-256",
        "microsoft/swinv2-base-patch4-window16-256",
        "microsoft/swinv2-base-patch4-window12-192-22k",
        "microsoft/swinv2-large-patch4-window12-192-22k",
        "microsoft/swinv2-base-patch4-window12to16-192to256-22kto1k-ft",
        "microsoft/swinv2-base-patch4-window12to24-192to384-22kto1k-ft",
        "microsoft/swinv2-large-patch4-window12to16-192to256-22kto1k-ft",
        "microsoft/swinv2-large-patch4-window12to24-192to384-22kto1k-ft",
    ],
    "BiT": ["google/bit-50"],
    "ViT": [
        "google/vit-base-patch16-224",
        "google/vit-base-patch16-224-in21k",
        "google/vit-base-patch16-384",
        "google/vit-base-patch32-384",
        "google/vit-base-patch32-224-in21k",
        "google/vit-large-patch16-224",
        "google/vit-large-patch16-224-in21k",
        "google/vit-large-patch16-384",
        "google/vit-large-patch32-384",
        "google/vit-large-patch32-224-in21k",
        #            'google/vit-huge-patch14-224-in21k',
    ],
    "DeiT": [
        "facebook/deit-tiny-patch16-224",
        "facebook/deit-small-patch16-224",
        "facebook/deit-base-patch16-224",
        "facebook/deit-base-patch16-384",
    ],
    "BEiT": [
        "microsoft/beit-base-patch16-224",
        "microsoft/beit-base-patch16-224-pt22k",
        "microsoft/beit-base-patch16-384",
        "microsoft/beit-large-patch16-224",
        "microsoft/beit-large-patch16-224-pt22k",
        "microsoft/beit-large-patch16-384",
        #                'microsoft/beit-large-patch16-512',
    ],
    "SwiftFormer": [
        "MBZUAI/swiftformer-xs",
        "MBZUAI/swiftformer-s",
        "MBZUAI/swiftformer-l1",
        "MBZUAI/swiftformer-l3",
    ],
    "DiNAT": [
        "shi-labs/dinat-tiny-in1k-224",
        "shi-labs/dinat-mini-in1k-224",
        "shi-labs/dinat-small-in1k-224",
        "shi-labs/dinat-base-in1k-224",
    ],
    "ResNet": [
        "microsoft/resnet-18",
        "microsoft/resnet-26",
        "microsoft/resnet-34",
        "microsoft/resnet-50",
        "microsoft/resnet-101",
        "microsoft/resnet-152",
    ],
}


class HuggingfaceClassifierTuner(BaseClassifierTuner):
    def __init__(
        self,
        base_dir,
        log_base_dir,
        dataset_name="composite_failures_1x",
        run_name_text="classifier-tuning",
        num_runs=1,
    ):
        super().__init__(base_dir, log_base_dir, dataset_name, run_name_text)
        self.num_runs = num_runs

    def _init_base_config(self, force_cpu=False):
        screenshot_height, screenshot_width = self._detect_screenshot_size()

        raw_model_names = []
        for model_class in hf_model_list.keys():
            for model_name in hf_model_list[model_class]:
                raw_model_names.append("{},{}".format(model_class, model_name))

        self.tunables = {
            "coded_model_name": raw_model_names,
            "num_epochs": [5, 10, 15],
            "batch_size": [4, 16, 32],
        }

        self.base_config = {
            "trainer_type": "huggingface",
            "image_height": screenshot_height,
            "image_width": screenshot_width,
            "force_cpu": force_cpu,
            "base_dir": self.base_dir,
            "random_seed": RANDOM_SEED,
            "mlflow_experiment_name": "OpenQA-rescheduler-pretrained-huggingface-{}".format(
                datetime.now().strftime(DATE_FORMAT)
            ),
            "dataset_path": self.dataset_path,
            "num_runs": self.num_runs,
        }

    def _generate_runs(self):
        tunable_lists = []

        for key in self.tunables.keys():
            tunable_list = []
            for param in self.tunables[key]:
                tunable_list.append("{}:{}:{}".format(key, param, type(param)))
            tunable_lists.append(tunable_list)

        for tunable_config in itertools.product(*tunable_lists):
            decoded_config = _decode_config(tunable_config)
            decoded_config.update(self.base_config)
            if "coded_model_name" in decoded_config.keys():
                model_class, model_name = decoded_config["coded_model_name"].split(",")
                decoded_config["model_class"] = model_class
                decoded_config["model_name"] = model_name
            self.configs.append(decoded_config)
