#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import click

from openqa_classifier.tuners import (
    custom_cnn,
    torch_pretrained,
    BaseClassifierTuner,
    huggingface_pretrained,
)


@click.command
@click.argument("trainer_name")
@click.option("--num-epochs", default=10, help="number of training epochs to use")
@click.option("--num-samples", default=10, help="number of training epochs to use")
@click.option(
    "--base-dir", default="../openqa_data/", help="base directory for the project"
)
@click.option(
    "--log-dir", default="./tuning_runs/", help="base directory log files generated"
)
@click.option(
    "--run-name-text",
    default="classifier-tuning-run",
    help="non-date part of the run name",
)
@click.option(
    "--dataset-name", default="subsampled_screenshots", help="name of dataset to use"
)
@click.option(
    "--force-cpu/--no-force-cpu",
    default=False,
    help="force use of CPU, even if CUDA is available",
)
@click.option(
    "--classifier-module",
    default="openqa_classifier.cli",
    help="python module to call to execute the classifier",
)
@click.option(
    "--num-runs",
    default=1,
    type=click.IntRange(1, 10),
    help="Number of times to run. >1 repeats identical run. only works with huggingface classifier",
)
def prep(
    num_epochs,
    num_samples,
    base_dir,
    log_dir,
    force_cpu,
    classifier_module,
    run_name_text,
    dataset_name,
    trainer_name,
    num_runs,
):
    if trainer_name == "cnn_3_layer":
        tuner = custom_cnn.CnnClassifierTuner(
            base_dir, log_dir, dataset_name=dataset_name, run_name_text=run_name_text
        )
    elif trainer_name == "torch":
        tuner = torch_pretrained.PretrainedClassifierTuner(
            base_dir, log_dir, dataset_name=dataset_name, run_name_text=run_name_text
        )
    elif trainer_name == "huggingface":
        tuner = huggingface_pretrained.HuggingfaceClassifierTuner(
            base_dir,
            log_dir,
            dataset_name=dataset_name,
            run_name_text=run_name_text,
            num_runs=num_runs,
        )
    else:
        raise Exception("invalid trainer name: {}".format())

    tuner.prep_tune(num_epochs, num_samples, classifier_module, force_cpu)


@click.argument("run_dir")
@click.option(
    "--classifier-module",
    default="openqa_classifier.classifier",
    help="python module to call to execute the classifier",
)
@click.option(
    "--min-cuda-memory",
    default=7000,
    help="minimum available CUDA memory before running",
)
@click.command
def run(run_dir, classifier_module, min_cuda_memory):
    tuner = BaseClassifierTuner("", "")
    tuner.run_prepped_tune(run_dir, classifier_module, min_cuda_memory=min_cuda_memory)


@click.group()
def cli():
    pass


cli.add_command(prep)
cli.add_command(run)

if __name__ == "__main__":
    cli()
