#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

"""
find_openqa_jobs.py

This script queries Fedora's datagrepper API to find completed OpenQA jobs and dumps those found jobs to a json file
for other job downloading script to use.

Basically, it generates the input needed by openqa_classifier.openqa to download OpenQA job data
"""

import argparse
from datetime import datetime, timedelta
import json
import os

import requests

import openqa_classifier.openqa

DATAGREPPER_URL = "https://apps.fedoraproject.org/datagrepper/v2/search"
OPENQA_TOPIC = "org.fedoraproject.prod.openqa.job.done"

TIME_FORMAT = "%Y%m%d-%H:%M:%S"
MAX_RETRIES = 3


def get_openqa_messages(starttime, endtime):

    total_pages = 1
    total_records = 0
    current_page = 1

    messages = []
    while current_page <= total_pages:
        num_retries = 0
        print("retrieving page {} of {}".format(current_page, total_pages))
        params = {
            "start": starttime,
            "end": endtime,
            "topic": OPENQA_TOPIC,
            "page": current_page,
        }

        while num_retries < MAX_RETRIES:
            try:
                r = requests.get(DATAGREPPER_URL, params=params)
                raw_data = r.json()
            except requests.RequestException as e:
                num_retries += 1
                print(
                    "Hit exception retrieving page {}, retrying ({} of {})".format(
                        current_page, num_retries, MAX_RETRIES
                    )
                )
                print(e)
            break

        total_records = raw_data["count"]
        total_pages = raw_data["pages"]

        messages.extend(raw_data["raw_messages"])
        current_page += 1

    return messages


def download_openqa_jobs(
    base_dir,
    output_filename,
    time_period,
    start_timestamp,
    end_timestamp,
    file_grouptime,
):
    """Remember, time_period, start_time or end_time can be None. time_period overrides start and end"""
    # yes, I know this is code duplication. don't care, it won't be used again (famous last words)

    if time_period is not None:
        endtime = datetime.utcnow().timestamp()
        starttime = endtime - time_period
    else:
        starttime = datetime.strptime(start_timestamp, TIME_FORMAT)
        endtime = datetime.strptime(end_timestamp, TIME_FORMAT)

    if file_grouptime is None:
        messages = get_openqa_messages(starttime, endtime)
        print("found {} messages, writing to single file".format(len(messages)))
        with open(output_filename, "w") as outputfile:
            for message in messages:
                outputfile.write(str(message["body"]["id"]))
                outputfile.write("\n")
    else:
        this_start = starttime
        file_num = 1

        # if there are existing files of the nameing scheme we're using, assume that we need to resume
        # again, I'm sure there are better ways to do this, like retry logic to handle the connection errors
        # *shrug*
        this_filename = "{}{}".format(output_filename, file_num)
        while os.path.isfile(this_filename):
            print("output file {} already exists, skipping".format(this_filename))
            file_num += 1
            this_filename = "{}{}".format(output_filename, file_num)
            this_start += timedelta(seconds=file_grouptime)

        print("starting at {}".format(this_start))
        print("using file {}".format(this_filename))

        while this_start < endtime:

            this_end = this_start + timedelta(seconds=file_grouptime)
            this_filename = "{}{}".format(output_filename, file_num)

            messages = get_openqa_messages(this_start, this_end)
            print(
                "found {} messages from {} to {}, writing to file {}".format(
                    len(messages), this_start, this_end, this_filename
                )
            )
            with open(this_filename, "w") as outputfile:
                for message in messages:
                    outputfile.write("{}\n".format(message["body"]["id"]))

            this_start = this_end
            file_num += 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="download old openqa job data")
    parser.add_argument(
        "-t",
        "--time_period",
        help="time period (in seconds) until now to use for downloading openqa job data (will override start and end times)",
        type=int,
    )
    parser.add_argument(
        "-s", "--start", help="start time to use (YYYYMMDD-HH:MM:SS assumed UTC)"
    )
    parser.add_argument(
        "-e", "--end", help="start time to use (YYYYMMDD-HH:MM:SS assumed UTC)"
    )
    parser.add_argument("-b", "--basedir", help="basedir to use", default=".")
    parser.add_argument(
        "-g",
        "--grouptime",
        help="create files with jobids from this number of seconds",
        type=int,
    )
    parser.add_argument("-o", "--outputfile", help="output file to use")

    args = parser.parse_args()

    base_dir = args.basedir
    output_filename = args.outputfile

    time_period = args.time_period
    start_time = args.start
    end_time = args.end
    group_time = args.grouptime

    print("time_period: {}".format(time_period))
    print("start_time: {}".format(start_time))
    print("end_time: {}".format(end_time))
    print("group_time: {}".format(group_time))
    download_openqa_jobs(
        base_dir, output_filename, time_period, start_time, end_time, group_time
    )
