#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import argparse
import math
import os

import click
from PIL import Image

# width and height of the input screenshots
IMAGE_WIDTH = 1024
IMAGE_HEIGHT = 768


# code duplication, also in prep_ocr_files and this stuff needs to be consolidated
def find_screenshot_files(base_dir, file_extension):
    all_screenshots = {}
    with os.scandir(base_dir) as jobdirscan:
        for entry in jobdirscan:
            if entry.is_dir():
                screenshot_dir = os.path.join(base_dir, entry.name, "screenshots")
            else:
                continue

            if not os.path.isdir(screenshot_dir):
                continue

            with os.scandir(screenshot_dir) as screenshotscan:
                screenshots = {}
                for screenshot_entry in screenshotscan:
                    if screenshot_entry.is_file():
                        # print("found file {}".format(screenshot_entry.name))
                        if screenshot_entry.name.endswith(file_extension):
                            # ew, this is complex - openqa names screenshots as [testname]-[number]
                            # and I want to group by the test name
                            testname = "-".join(screenshot_entry.name.split("-")[:-1])

                            # sometimes, we end up with a category that is empty, that's not what we want
                            if len(testname) == 0:
                                continue

                            if not testname in screenshots.keys():
                                screenshots[testname] = [screenshot_entry.name]
                            else:
                                screenshots[testname].append(screenshot_entry.name)

                all_screenshots[entry.name] = screenshots
    return all_screenshots


def make_composite_screenshots(png_files, base_dir, output_dir, scale):

    if not os.path.isdir(output_dir):
        print("Creating output directory: {}".format(output_dir))
        os.makedirs(output_dir)

    for jobid in png_files:
        print("job {}".format(jobid))

        for category in png_files[jobid].keys():

            num_screenshots = len(png_files[jobid][category])
            print("  {} ({} screenshots)".format(category, num_screenshots))

            # find the number of screenshots on each side of the composite image
            composite_len = math.ceil(math.sqrt(num_screenshots))

            composite_width = composite_len * IMAGE_WIDTH
            composite_height = composite_len * IMAGE_HEIGHT
            composite_image = Image.new(
                "RGB", (composite_width, composite_height), color=(255, 255, 255)
            )

            column_i = 0
            row_i = 0

            for png_file in png_files[jobid][category]:

                image_filename = os.path.join(base_dir, jobid, "screenshots", png_file)
                pil_image = Image.open(image_filename)
                cpil_image = pil_image.convert("RGB")

                start_width = column_i * IMAGE_WIDTH
                start_height = row_i * IMAGE_HEIGHT

                composite_image.paste(cpil_image, (start_width, start_height))

                # handle munging of the i variables
                if column_i >= composite_len:
                    row_i += 1
                    column_i = 0
                else:
                    column_i += 1

            resized_height = math.floor(scale * IMAGE_HEIGHT)
            resized_width = math.floor(scale * IMAGE_WIDTH)

            resized_composite = composite_image.resize((resized_width, resized_height))
            composite_filename = os.path.join(
                output_dir, "{}.{}.composite.png".format(jobid, category)
            )
            resized_composite.save(composite_filename)


@click.command
@click.option(
    "--base-dir",
    default="../openqa_data/raw_data",
    help="base directory to look for input in",
)
@click.option(
    "--output-dir",
    default="../openqa_data/composite_screenshots",
    help="directory into which composites should be written",
)
@click.option(
    "--composite-scale",
    default=2,
    help="scale of the composite screenshot, relative to the input screenshots",
)
def cli(base_dir, output_dir, composite_scale):
    png_files = find_screenshot_files(base_dir, ".png")
    print("found {} png files".format(len(png_files)))

    make_composite_screenshots(png_files, base_dir, output_dir, scale=composite_scale)


if __name__ == "__main__":
    cli()
