#  -*- coding: utf-8 -*-
# Copyright (c) 2023 Timothy Flink.
# License: GPL-3.0-or-later
# See the LICENSE file for more details on licensing
#
# This file is part of openqa_classifier
# (see https://pagure.io/fedora-qa/openqa_classifier).
#

import copy
import json
import os

import click
import pandas


def make_openqa_link(job_id):
    return "https://openqa.fedoraproject.org/tests/{}".format(job_id)


@click.command()
@click.argument("job_dir")
@click.option("-o", "--outputfile", default="openqa_api_data.csv", help="filename for output api data")
def gather_and_dump_api_data(job_dir, outputfile):
    all_data = {}
    # find all the json files
    with os.scandir(job_dir) as base_scan:
        for entry in base_scan:
            if entry.is_dir():
                job_id = int(entry.name)
                # print("Job {}:".format(job_id))
                # print("  link: {}".format(make_openqa_link(job_id)))
                json_filename = "{}{}/{}_apidata.json".format(job_dir, job_id, job_id)
                if not os.path.exists(json_filename):
                    print("  Missing API json file for {}".format(job_id))
                    # print()
                    continue
                with open(json_filename, "r") as json_file:
                    json_data = json.loads(json_file.read())

                raw_job_data = json_data["jobs"][0]
                job_data = {
                    "job_id": job_id,
                    "href": make_openqa_link(job_id),
                    "rescheduled_as": -1,
                }

                # look for the origin_id key
                if "origin_id" in raw_job_data.keys():
                    job_data["origin_id"] = int(raw_job_data["origin_id"])
                else:
                    job_data["origin_id"] = -1

                # check for pass/fail
                job_data["overall_result"] = raw_job_data["result"]

                job_data["test"] = raw_job_data["test"]
                job_data["t_started"] = raw_job_data["t_started"]
                job_data["t_finished"] = raw_job_data["t_finished"]
                job_data["arch"] = raw_job_data["settings"]["ARCH"]
                job_data["version"] = raw_job_data["settings"]["VERSION"]

                job_data["modules"] = []
                # loop through the modules
                for module in raw_job_data["modules"]:
                    module_data = {"module": module["name"], "result": module["result"]}
                    job_data["modules"].append(module_data)

                all_data[job_id] = job_data

    # mark all rescheduled jobs
    for jobid in all_data.keys():
        job_originid = all_data[jobid]["origin_id"]
        if job_originid != -1:
            print("job {} rescheduled as {}".format(job_originid, jobid))
            if job_originid in all_data:
                # out of curiosity, can you have a job rescheduled multiple times?
                if all_data[job_originid]["rescheduled_as"] != -1:
                    print("Already rescheduled job {} as {}".format(job_originid, jobid))
                all_data[job_originid]["rescheduled_as"] = jobid

    # flatten the job data
    flat_data = []
    for jobid in all_data.keys():
        job_data = all_data[jobid]
        for module in job_data["modules"]:
            module_data = copy.deepcopy(job_data)
            module_data["module"] = module["module"]
            module_data["module_result"] = module["result"]
            flat_data.append(module_data)

    columns = [
        "job_id",
        "test",
        "overall_result",
        "module",
        "module_result",
        "rescheduled_as",
        "origin_id",
        "t_started",
        "t_finished",
        "arch",
        "version",
        "href",
    ]
    df = pandas.DataFrame(flat_data, columns=columns)

    df.to_csv(outputfile)


if __name__ == "__main__":
    gather_and_dump_api_data()
